from typing import Type, Any, Optional, overload

from .core import resource as res
from .core.client import Client
from .core.async_client import AsyncClient
from .core.internal_models import meta_v1, autoscaling_v1
from .core.internal_resources import apiextensions
from .core.resource_registry import resource_registry


__all__ = [
    "async_load_in_cluster_generic_resources",
    "create_global_resource",
    "create_namespaced_resource",
    "create_resources_from_crd",
    "get_generic_resource",
    "load_in_cluster_generic_resources",
]


def get_generic_resource(version, kind):
    """Query generic resources already defined using one of the other methods described in this module or via
    `codecs.load_all_yaml(..., create_resources_for_crds=True)`

    **Parameters**

    * **version** `str` - Resource version including the API group. Example `stable.example.com/v1`
    * **kind** `str` - Resource kind. Example: `CronTab`

    **returns** class representing the generic resource or `None` if it's not found
    """
    resource = resource_registry.get(version, kind)
    if resource is None or not issubclass(
        resource, (GenericGlobalResource, GenericNamespacedResource)
    ):
        return None
    return resource


class Generic(dict):
    @overload
    def __init__(
        self,
        apiVersion: str = None,
        kind: str = None,
        metadata: meta_v1.ObjectMeta = None,
        **kwargs,
    ):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def apiVersion(self) -> str:
        return self.get("apiVersion")

    @property
    def kind(self) -> str:
        return self.get("kind")

    @property
    def status(self) -> str:
        return self.get("status")

    @property
    def metadata(self) -> Optional[meta_v1.ObjectMeta]:
        meta = self.get("metadata")
        if meta is None:
            return None
        elif isinstance(meta, meta_v1.ObjectMeta):
            return meta
        return meta_v1.ObjectMeta.from_dict(meta)

    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(f"{item} not found")
        return self.get(item)

    @classmethod
    def from_dict(cls, d: dict, lazy=True):
        return cls(d)

    def to_dict(self, dict_factory=dict):
        d = dict_factory(self)
        if "metadata" in d and isinstance(d["metadata"], meta_v1.ObjectMeta):
            d["metadata"] = d["metadata"].to_dict(dict_factory)
        return d


def create_api_info(group, version, kind, plural, verbs=None) -> res.ApiInfo:
    if verbs is None:
        verbs = [
            "delete",
            "deletecollection",
            "get",
            "global_list",
            "global_watch",
            "list",
            "patch",
            "post",
            "put",
            "watch",
        ]
    return res.ApiInfo(
        resource=res.ResourceDef(group, version, kind), plural=plural, verbs=verbs
    )


class GenericGlobalScale(res.GlobalSubResource, autoscaling_v1.Scale):
    pass


class GenericGlobalStatus(res.GlobalSubResource, Generic):
    pass


class GenericNamespacedScale(res.NamespacedResourceG, autoscaling_v1.Scale):
    pass


class GenericNamespacedStatus(res.NamespacedResourceG, Generic):
    pass


def _create_subresource(main_class, parent_info: res.ApiInfo, action):
    class TmpName(main_class):
        _api_info = res.ApiInfo(
            resource=(
                parent_info.resource
                if action == "status"
                else res.ResourceDef("autoscaling", "v1", "Scale")
            ),
            parent=parent_info.resource,
            plural=parent_info.plural,
            verbs=["get", "patch", "put"],
            action=action,
        )

    TmpName.__name__ = TmpName.__qualname__ = (
        f"{parent_info.resource.kind}{action.capitalize()}"
    )
    return TmpName


class GenericGlobalResource(res.GlobalResource, Generic):
    Scale: Type[GenericGlobalScale]
    Status: Type[GenericGlobalStatus]


class GenericNamespacedResource(res.NamespacedResourceG, Generic):
    Scale: Type[GenericNamespacedScale]
    Status: Type[GenericNamespacedStatus]


def _api_info_signature(api_info: res.ApiInfo, namespaced: bool):
    return (
        namespaced,
        api_info.plural,
        tuple(api_info.verbs) if api_info.verbs else None,
    )


def _create_resource(namespaced, group, version, kind, plural, verbs=None) -> Any:
    model = resource_registry.get(f"{group}/{version}", kind)
    api_info = create_api_info(group, version, kind, plural, verbs=verbs)
    signature = _api_info_signature(api_info, namespaced)

    if model is not None:
        curr_namespaced = issubclass(model, res.NamespacedResource)
        curr_signature = _api_info_signature(model._api_info, curr_namespaced)
        if curr_signature != signature:
            raise ValueError(
                f"Resource {kind} already created but with different signature"
            )
        return model

    if namespaced:
        main, status, scale = (
            GenericNamespacedResource,
            GenericNamespacedStatus,
            GenericNamespacedScale,
        )
    else:
        main, status, scale = (
            GenericGlobalResource,
            GenericGlobalStatus,
            GenericGlobalScale,
        )

    class TmpName(main):
        _api_info = create_api_info(group, version, kind, plural, verbs=verbs)

        Scale = _create_subresource(scale, _api_info, action="scale")
        Status = _create_subresource(status, _api_info, action="status")

    TmpName.__name__ = TmpName.__qualname__ = kind
    return resource_registry.register(TmpName)


def create_global_resource(
    group: str, version: str, kind: str, plural: str, verbs=None
) -> Type[GenericGlobalResource]:
    """Create a new class representing a global resource with the provided specifications.

    **Parameters**

    * **group** `str` - API group of the resource. Example `stable.example.com`.
    * **version** `str` - API group version. Example `v1`.
    * **kind** `str` - Resource name. Example `Job`.
    * **plural** `str` - Resource collection name. Example `jobs`.

    **returns**  Subclass of `GenericGlobalResource`.
    """
    return _create_resource(False, group, version, kind, plural, verbs=verbs)


def create_namespaced_resource(
    group: str, version: str, kind: str, plural: str, verbs=None
) -> Type[GenericNamespacedResource]:
    """Create a new class representing a namespaced resource with the provided specifications.

    **Parameters**

    * **group** `str` - API group of the resource. Example `stable.example.com`.
    * **version** `str` - API group version. Example `v1`.
    * **kind** `str` - Resource name. Example `Job`.
    * **plural** `str` - Resource collection name. Example `jobs`.

    **returns**  Subclass of `GenericNamespacedResource`.
    """
    return _create_resource(True, group, version, kind, plural, verbs=verbs)


def load_in_cluster_generic_resources(client: Client):
    """Loads all in-cluster CustomResourceDefinitions as generic resources.

    Once loaded, generic resources can be obtained from `generic_resource.get_generic_resource()`,
    or used implicitly such as when using `codecs.load_all_yaml()`.

    **Parameters**

    * **client** `Client` - Lightkube Client to use to load the CRDs.
    """
    crds = client.list(apiextensions.CustomResourceDefinition)
    for crd in crds:
        create_resources_from_crd(crd)


async def async_load_in_cluster_generic_resources(client: AsyncClient):
    """Loads all in-cluster CustomResourceDefinitions as generic resources.

    Once loaded, generic resources can be obtained from `generic_resource.get_generic_resource()`,
    or used implicitly such as when using `codecs.load_all_yaml()`.

    **Parameters**

    * **client** `AsyncClient` - Lightkube AsyncClient to use to load the CRDs.
    """
    crds = client.list(apiextensions.CustomResourceDefinition)
    async for crd in crds:
        create_resources_from_crd(crd)


def create_resources_from_crd(crd: apiextensions.CustomResourceDefinition):
    """Creates a generic resource for each version in a CustomResourceDefinition."""
    if crd.spec.scope == "Namespaced":
        creator = create_namespaced_resource
    elif crd.spec.scope == "Cluster":
        creator = create_global_resource
    else:
        raise ValueError(
            f"Unexpected scope for resource.  Expected 'Namespaced' or 'Cluster',"
            f" got {crd.spec.scope}"
        )

    for version in crd.spec.versions:
        creator(**_crd_to_dict(crd, version.name))


def _crd_to_dict(crd, version_name):
    return {
        "group": crd.spec.group,
        "version": version_name,
        "kind": crd.spec.names.kind,
        "plural": crd.spec.names.plural,
    }



import pytest
from unittest import mock

from lightkube import generic_resource as gr
from lightkube.core.generic_client import GenericClient
from lightkube.models.meta_v1 import ObjectMeta
from lightkube.resources.apiextensions_v1 import CustomResourceDefinition
from lightkube.models.apiextensions_v1 import (
    CustomResourceDefinitionNames,
    CustomResourceDefinitionSpec,
    CustomResourceDefinitionVersion,
)
from lightkube.core.resource_registry import resource_registry


def create_dummy_crd(group="thisgroup", kind="thiskind", plural="thiskinds", scope="Namespaced",
                     versions=None):
    if versions is None:
        versions = ['v1alpha1', 'v1']

    crd = CustomResourceDefinition(
        spec=CustomResourceDefinitionSpec(
            group=group,
            names=CustomResourceDefinitionNames(
                kind=kind,
                plural=plural,
            ),
            scope=scope,
            versions=[
                CustomResourceDefinitionVersion(
                    name=version,
                    served=True,
                    storage=True,
                ) for version in versions
            ],
        )
    )

    return crd


@pytest.fixture(autouse=True)
def cleanup_registry():
    """Cleanup the registry before each test"""
    yield
    resource_registry.clear()


class MockedClient(GenericClient):
    def __init__(self):
        self.namespace = 'default'
        self._field_manager = None
        self._dry_run = False


@pytest.fixture()
def mocked_client_list_crds():
    """Yields a Client with a mocked .list which returns a fixed list of CRDs

    **returns**  Tuple of: mocked `Client`, list of CRDs, integer number of resources defined by
                 CRDs
    """
    scopes = ["Namespaced", "Cluster"]
    version_names = ['v2', 'v3']

    crds = [create_dummy_crd(scope=scope, kind=scope, versions=version_names) for scope in scopes]
    expected_n_resources = len(version_names) * len(crds)

    with mock.patch("lightkube.Client") as client_maker:
        mocked_client = mock.MagicMock()
        mocked_client.list.return_value = crds
        client_maker.return_value = mocked_client
        yield mocked_client, crds, expected_n_resources


class AsyncIterator:
    """Provides a `async for` compatible iterator

    Pattern taken from https://stackoverflow.com/a/36724229/5394584
    """
    def __init__(self, seq):
        self.iter = iter(seq)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.iter)
        except StopIteration:
            raise StopAsyncIteration


@pytest.mark.asyncio
@pytest.fixture()
def mocked_asyncclient_list_crds():
    """Yields an AsyncClient with a mocked .list which returns a fixed list of CRDs

    **returns**  Tuple of: mocked `AsyncClient`, list of CRDs, integer number of resources defined by
                 CRDs
    """
    scopes = ["Namespaced", "Cluster"]
    version_names = ['v2', 'v3']

    crds = [create_dummy_crd(scope=scope, kind=scope, versions=version_names) for scope in scopes]
    asynccrds = AsyncIterator(crds)
    expected_n_resources = len(version_names) * len(crds)

    with mock.patch("lightkube.AsyncClient") as client_maker:
        # This can be removed when python < 3.8 is not supported
        try:
            mocked_client = mock.AsyncMock()
        except AttributeError:
            import asyncmock
            mocked_client = asyncmock.AsyncMock()

        # AsyncClient.list is not async, but AsyncMock will automatically generate it as async.
        # Instead, mock it explicitly with a regular MagicMock
        mocked_list = mock.MagicMock()
        mocked_list.return_value = asynccrds
        mocked_client.list = mocked_list
        client_maker.return_value = mocked_client
        yield mocked_client, crds, expected_n_resources


def test_create_namespaced_resource():
    c = MockedClient()
    Test = gr.create_namespaced_resource('test.eu', 'v1', 'TestN', 'tests')
    assert Test.__name__ == 'TestN'

    pr = c.prepare_request('get', Test, name='xx', namespace='myns')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/namespaces/myns/tests/xx'

    pr = c.prepare_request('list', Test, namespace='myns')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/namespaces/myns/tests'

    pr = c.prepare_request('post', obj=Test(metadata={'namespace': 'myns'}, spec={'a': 1}))
    assert pr.method == 'POST'
    assert pr.url == 'apis/test.eu/v1/namespaces/myns/tests'
    assert pr.data == {'apiVersion': 'test.eu/v1', 'kind': 'TestN', 'spec': {'a': 1}, 'metadata': {'namespace': 'myns'}}

    pr = c.prepare_request('get', Test.Scale, name='xx', namespace='myns')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/namespaces/myns/tests/xx/scale'

    pr = c.prepare_request('get', Test.Status, name='xx', namespace='myns')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/namespaces/myns/tests/xx/status'


def test_create_global_resource():
    c = MockedClient()
    Test = gr.create_global_resource('test.eu', 'v1', 'TestG', 'tests')
    assert Test.__name__ == 'TestG'

    pr = c.prepare_request('get', Test, name='xx')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/tests/xx'

    pr = c.prepare_request('list', Test)
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/tests'

    pr = c.prepare_request('post', obj=Test(spec={'a': 1}))
    assert pr.method == 'POST'
    assert pr.url == 'apis/test.eu/v1/tests'
    assert pr.data == {'apiVersion': 'test.eu/v1', 'kind': 'TestG', 'spec': {'a': 1}}

    pr = c.prepare_request('get', Test.Scale, name='xx')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/tests/xx/scale'

    pr = c.prepare_request('get', Test.Status, name='xx')
    assert pr.method == 'GET'
    assert pr.url == 'apis/test.eu/v1/tests/xx/status'


@pytest.mark.parametrize(
    "crd_scope",
    [
        "Namespaced",
        "Cluster",
    ]
)
def test_create_resources_from_crd(crd_scope):
    version_names = ['v1alpha1', 'v1', 'v2']
    crd = create_dummy_crd(scope=crd_scope, versions=version_names)

    # Confirm no generic resources exist before testing
    assert len(resource_registry._registry) == 0

    # Test the function
    gr.create_resources_from_crd(crd)

    # Confirm expected number of resources created
    assert len(resource_registry._registry) == len(version_names)

    # Confirm expected resources exist
    for version in version_names:
        resource = gr.get_generic_resource(f"{crd.spec.group}/{version}", crd.spec.names.kind)
        assert resource is not None


def test_generic_model():
    mod = gr.Generic.from_dict({'metadata': {'name': 'bla'}, 'test': {'ok': 4}})
    assert mod.metadata.name == 'bla'
    assert mod.test['ok'] == 4
    assert mod.to_dict() == {'metadata': {'name': 'bla'}, 'test': {'ok': 4}}
    assert mod.status is None

    mod = gr.Generic.from_dict({'apiVersion': 'v1', 'kind': 'Test', 'status': 1})
    assert mod.apiVersion == 'v1'
    assert mod.kind == 'Test'
    assert mod.metadata is None
    assert mod.to_dict() == {'apiVersion': 'v1', 'kind': 'Test', 'status': 1}
    assert mod.status == 1

    mod = gr.Generic(metadata=ObjectMeta(name='bla'), test={'ok': 4})
    assert mod.metadata.name == 'bla'
    assert mod.test['ok'] == 4
    assert mod.to_dict() == {'metadata': {'name': 'bla'}, 'test': {'ok': 4}}
    assert mod.status is None

    with pytest.raises(AttributeError):
        mod._a


def test_load_in_cluster_generic_resources(mocked_client_list_crds):
    """Test that load_in_cluster_generic_resources creates generic resources for crds in cluster"""
    # Set up environment
    mocked_client, expected_crds, expected_n_resources = mocked_client_list_crds

    # Confirm no generic resources exist before testing
    assert len(resource_registry._registry) == 0

    # Test the function
    gr.load_in_cluster_generic_resources(mocked_client)

    # Confirm the expected resources and no others were created
    assert len(resource_registry._registry) == expected_n_resources
    for crd in expected_crds:
        for version in crd.spec.versions:
            resource = gr.get_generic_resource(f"{crd.spec.group}/{version.name}", crd.spec.names.kind)
            assert resource is not None

    mocked_client.list.assert_called_once()


@pytest.mark.asyncio
async def test_async_load_in_cluster_generic_resources(mocked_asyncclient_list_crds):
    """Test that async_load_in_cluster_generic_resources creates generic resources for crds in cluster"""
    # Set up environment
    mocked_client, expected_crds, expected_n_resources = mocked_asyncclient_list_crds

    # Confirm no generic resources exist before testing
    assert len(resource_registry._registry) == 0

    # Test the function
    await gr.async_load_in_cluster_generic_resources(mocked_client)

    # Confirm the expected resources and no others were created
    assert len(resource_registry._registry) == expected_n_resources
    for crd in expected_crds:
        for version in crd.spec.versions:
            resource = gr.get_generic_resource(f"{crd.spec.group}/{version.name}", crd.spec.names.kind)
            assert resource is not None

    # This only works for python >3.8, not for the asyncmock package needed in <3.8
    # mocked_client.list.assert_called_once()


def test_scale_model():
    """Test we are using the right model here"""
    Test = gr.create_global_resource('test.eu', 'v1', 'TestS', 'tests')
    a = Test.Scale.from_dict({'spec': {'replicas': 2}})
    assert a.spec.replicas == 2


def test_signature_change_not_allowed():
    gr.create_namespaced_resource('test.eu', 'v1', 'TestN', 'tests')
    gr.create_namespaced_resource('test.eu', 'v1', 'TestN', 'tests')

    with pytest.raises(ValueError, match='.*different signature'):
        gr.create_namespaced_resource('test.eu', 'v1', 'TestN', 'tests', verbs=['get'])

    with pytest.raises(ValueError, match='.*different signature'):
        gr.create_global_resource('test.eu', 'v1', 'TestN', 'tests')



from typing import Dict, List
from dataclasses import dataclass, field
import tempfile
import base64

from ..core.dataclasses_dict import DataclassDictMixIn


class FileStr(str):
    def __new__(cls, data):
        if data is None:
            return None

        f = tempfile.NamedTemporaryFile()
        f.write(base64.b64decode(data))
        f.flush()
        file = str.__new__(cls, f.name)
        file.handler = f
        return file

    def __del__(self):
        if self.handler:
            self.handler.close()
            self.handler = None


@dataclass
class Context(DataclassDictMixIn):
    cluster: str
    user: str = None
    namespace: str = None


@dataclass
class NameValue(DataclassDictMixIn):
    name: str
    value: str


@dataclass
class UserExec(DataclassDictMixIn):
    apiVersion: str
    command: str = None
    env: List[NameValue] = field(default_factory=list)
    args: List[str] = field(default_factory=list)
    installHint: str = None


@dataclass
class User(DataclassDictMixIn):
    exec: UserExec = None
    username: str = None
    password: str = None
    token: str = None
    auth_provider: Dict = field(metadata={"json": "auth-provider"}, default=None)
    client_cert: str = field(metadata={"json": "client-certificate"}, default=None)
    client_cert_data: str = field(
        metadata={"json": "client-certificate-data"}, default=None
    )
    client_key: str = field(metadata={"json": "client-key"}, default=None)
    client_key_data: str = field(metadata={"json": "client-key-data"}, default=None)


@dataclass
class Cluster(DataclassDictMixIn):
    server: str = "http://localhost:8080"
    certificate_auth: str = field(
        metadata={"json": "certificate-authority"}, default=None
    )
    certificate_auth_data: str = field(
        metadata={"json": "certificate-authority-data"}, default=None
    )
    insecure: bool = field(metadata={"json": "insecure-skip-tls-verify"}, default=False)



from typing import List
from ..core import resource as r


def sort_objects(
    objs: List[r.Resource], by: str = "kind", reverse: bool = False
) -> List[r.Resource]:
    """Sorts a list of resource objects by a sorting schema, returning a new list

    **parameters**

    * **objs** - list of resource objects to be sorted
    * **by** - *(optional)* sorting schema. Possible values:
        * `'kind'` - sorts by kind, ranking objects in an order that is suitable for batch-applying
          many resources.  For example, Namespaces and ServiceAccounts are sorted ahead of
          ClusterRoleBindings or Pods that might use them.  The reverse of this order is suitable
          for batch-deleting.
          See _kind_rank_function for full details on sorting
    * **reverse** - *(optional)* if `True`, sorts in reverse order
    """
    if by == "kind":
        objs = sorted(objs, key=_kind_rank_function, reverse=reverse)
    else:
        raise ValueError(f"Unknown sorting schema: {by}")
    return objs


UNKNOWN_ITEM_SORT_VALUE = 1000
RANK_ORDER = {
    "CustomResourceDefinition": 10,
    "Namespace": 20,
    "Secret": 31,
    "ServiceAccount": 32,
    "PersistentVolume": 33,
    "PersistentVolumeClaim": 34,
    "ConfigMap": 35,
    "Role": 41,
    "ClusterRole": 42,
    "RoleBinding": 43,
    "ClusterRoleBinding": 44,
}


def _kind_rank_function(obj: List[r.Resource]) -> int:
    """
    Returns an integer rank based on an objects .kind

    Ranking is set to order kinds by:
    * CRDs
    * Namespaces
    * Things that might be referenced by pods (Secret, ServiceAccount, PVs/PVCs, ConfigMap)
    * RBAC
        * Roles and ClusterRoles
        * RoleBindings and ClusterRoleBindings
    * Everything else (Pod, Deployment, ...)
    """
    return RANK_ORDER.get(obj.kind, UNKNOWN_ITEM_SORT_VALUE)



import time
from typing import (
    AsyncIterable,
    Type,
    Any,
    Dict,
    Union,
    Iterator,
    AsyncIterator,
    Tuple,
    TypeVar,
    Iterable,
    Optional,
)
import dataclasses
from dataclasses import dataclass
import json
import asyncio

import httpx

from . import resource as r
from ..config.kubeconfig import KubeConfig, SingleConfig, DEFAULT_KUBECONFIG
from ..config import client_adapter
from .exceptions import ApiError, NotReadyError
from ..types import OnErrorAction, OnErrorHandler, on_error_raise, PatchType


ALL_NS = "*"


def transform_exception(e: httpx.HTTPError):
    if (
        isinstance(e, httpx.HTTPStatusError)
        and e.response.headers["Content-Type"] == "application/json"
    ):
        return ApiError(request=e.request, response=e.response)
    return e


METHOD_MAPPING = {
    "delete": "DELETE",
    "deletecollection": "DELETE",
    "get": "GET",
    "global_list": "GET",
    "global_watch": "GET",
    "list": "GET",
    "patch": "PATCH",
    "post": "POST",
    "put": "PUT",
    "watch": "GET",
}


@dataclass
class BasicRequest:
    method: str
    url: str
    response_type: Any
    params: Dict[str, str] = dataclasses.field(default_factory=dict)
    data: Any = None
    headers: Dict[str, str] = None


class WatchDriver:
    def __init__(self, br: BasicRequest, build_request, lazy):
        self._version = br.params.get("resourceVersion")
        self._convert = br.response_type.from_dict
        self._br = br
        self._build_request = build_request
        self._lazy = lazy

    def get_request(self, timeout):
        br = self._br
        if self._version is not None:
            br.params["resourceVersion"] = self._version
        return self._build_request(br.method, br.url, params=br.params, timeout=timeout)

    def process_one_line(self, line):
        line = json.loads(line)
        tp = line["type"]
        obj = line["object"]
        self._version = obj["metadata"]["resourceVersion"]
        return tp, self._convert(obj, lazy=self._lazy)


T = TypeVar("T")


class ListIterable(Iterable[T]):

    _resourceVersion: Optional[str] = None

    @property
    def resourceVersion(self) -> str:
        """Returns the resource version at which the collection was constructed.
        Note: this property is only available after the iteration started and will raise NotReadyError otherwise
        """
        if self._resourceVersion:
            return self._resourceVersion
        raise NotReadyError(
            "resourceVersion", "only available after the iteration started"
        )

    def __init__(self, inner_iter: Iterator[Tuple[str, Iterator[T]]]) -> None:
        self._inner_iter = inner_iter

    def __iter__(self) -> Iterator[T]:
        for rv, chunk in self._inner_iter:
            self._resourceVersion = rv
            yield from chunk


class ListAsyncIterable(AsyncIterable[T]):

    _resourceVersion: Optional[str] = None

    @property
    def resourceVersion(self) -> str:
        """Returns the resource version at which the collection was constructed.
        Note: this property is only available after the iteration started and will raise NotReadyError otherwise
        """
        if self._resourceVersion:
            return self._resourceVersion
        raise NotReadyError(
            "resourceVersion", "only available after the iteration started"
        )

    def __init__(self, inner_iter: AsyncIterator[Tuple[str, Iterator[T]]]) -> None:
        self._inner_iter = inner_iter

    async def __aiter__(self) -> AsyncIterator[T]:
        async for rv, chunk in self._inner_iter:
            self._resourceVersion = rv
            for item in chunk:
                yield item


class GenericClient:
    AdapterClient = staticmethod(client_adapter.Client)

    def __init__(
        self,
        config: Union[SingleConfig, KubeConfig] = None,
        namespace: str = None,
        timeout: httpx.Timeout = None,
        lazy=True,
        trust_env: bool = True,
        field_manager: str = None,
        dry_run: bool = False,
        transport: Union[httpx.BaseTransport, httpx.AsyncBaseTransport] = None,
    ):
        self._timeout = httpx.Timeout(10) if timeout is None else timeout
        self._watch_timeout = httpx.Timeout(self._timeout)
        self._watch_timeout.read = None
        self._lazy = lazy
        if config is None and trust_env:
            config = KubeConfig.from_env().get()
        elif config is None and not trust_env:
            config = KubeConfig.from_file(DEFAULT_KUBECONFIG).get()
        elif isinstance(config, KubeConfig):
            config = config.get()

        self.config = config
        self._client = self.AdapterClient(
            config, timeout, trust_env=trust_env, transport=transport
        )
        self._field_manager = field_manager
        self._dry_run = dry_run
        self.namespace = namespace if namespace else config.namespace

    def prepare_request(
        self,
        method,
        res: Type[r.Resource] = None,
        obj=None,
        name=None,
        namespace=None,
        watch: bool = False,
        params: dict = None,
        headers: dict = None,
    ) -> BasicRequest:
        if params is not None:
            params = {k: v for k, v in params.items() if v is not None}
        else:
            params = {}
        if headers is not None:
            headers = {k: v for k, v in headers.items() if v is not None}
        data = None
        if res is None:
            if obj is None:
                raise ValueError(
                    "At least a resource or an instance of a resource need to be provided"
                )
            res = obj.__class__

        namespaced = issubclass(res, (r.NamespacedResource, r.NamespacedSubResource))

        if namespace == ALL_NS:
            if not issubclass(res, r.NamespacedResourceG):
                raise ValueError(f"Class {res} doesn't support global {method}")
            if method not in ("list", "watch"):
                raise ValueError(
                    "Only methods 'list' and 'watch' can be called for all namespaces"
                )
            real_method = "global_watch" if watch else "global_" + method
        else:
            real_method = "watch" if watch else method

        api_info = r.api_info(res)
        if real_method not in api_info.verbs:
            if watch:
                raise ValueError(f"Resource '{res.__name__}' is not watchable")
            else:
                raise ValueError(
                    f"method '{method}' not supported by resource '{res.__name__}'"
                )

        if watch:
            params["watch"] = "true"

        if api_info.parent is None:
            base = api_info.resource
        else:
            base = api_info.parent

        if base.group == "":
            path = ["api", base.version]
        else:
            path = ["apis", base.group, base.version]

        if namespaced and namespace != ALL_NS:
            if method in ("post", "put") and obj.metadata.namespace is not None:
                if namespace is None:
                    namespace = obj.metadata.namespace
                elif namespace != obj.metadata.namespace:
                    raise ValueError(
                        f"The namespace value '{namespace}' differ from the "
                        f"namespace in the object metadata '{obj.metadata.namespace}'"
                    )
            if namespace is None:
                namespace = self.namespace
            path.extend(["namespaces", namespace])

        if method in ("post", "put", "patch"):
            if self._field_manager is not None and "fieldManager" not in params:
                params["fieldManager"] = self._field_manager
            if self._dry_run is True and "dryRun" not in params:
                params["dryRun"] = "All"
            if (
                method == "patch"
                and headers["Content-Type"] == PatchType.APPLY.value
                and "fieldManager" not in params
            ):
                raise ValueError(
                    'Parameter "field_manager" is required for PatchType.APPLY'
                )
            if obj is None:
                raise ValueError("obj is required for post, put or patch")

            if method == "patch" and not isinstance(obj, r.Resource):
                data = obj
            else:
                data = obj.to_dict()
                # The following block, ensures that apiVersion and kind are always set.
                # this is needed as k8s fails if this data are not provided for objects derived by CRDs (Issue #27)
                if "apiVersion" not in data:
                    data["apiVersion"] = api_info.resource.api_version
                if "kind" not in data:
                    data["kind"] = api_info.resource.kind

        path.append(api_info.plural)
        if method in ("delete", "get", "patch", "put") or api_info.action:
            if name is None and method == "put":
                name = obj.metadata.name
            if name is None:
                raise ValueError("resource name not defined")
            path.append(name)

        if api_info.action:
            path.append(api_info.action)

        http_method = METHOD_MAPPING[method]
        if http_method == "DELETE":
            res = None

        return BasicRequest(
            method=http_method,
            url="/".join(path),
            params=params,
            response_type=res,
            data=data,
            headers=headers,
        )

    @staticmethod
    def raise_for_status(resp):
        try:
            resp.raise_for_status()
        except httpx.HTTPError as e:
            raise transform_exception(e)

    def build_adapter_request(self, br: BasicRequest):
        return self._client.build_request(
            br.method, br.url, params=br.params, json=br.data, headers=br.headers
        )

    def convert_to_resource(self, res: Type[r.Resource], item: dict) -> r.Resource:
        resource_def = r.api_info(res).resource
        item.setdefault("apiVersion", resource_def.api_version)
        item.setdefault("kind", resource_def.kind)
        return res.from_dict(item, lazy=self._lazy)

    def handle_response(self, method, resp, br):
        self.raise_for_status(resp)
        res = br.response_type
        if res is None:
            # TODO: delete/deletecollection actions normally return a Status object, we may want to return it as well
            return
        data = resp.json()
        if method == "list":
            if "metadata" in data and data["metadata"].get("continue"):
                cont = True
                br.params["continue"] = data["metadata"]["continue"]
            else:
                cont = False
            try:
                rv = data["metadata"]["resourceVersion"]
            except KeyError:
                rv = None
            return (
                cont,
                rv,
                (self.convert_to_resource(res, obj) for obj in data["items"]),
            )
        else:
            if res is not None:
                return self.convert_to_resource(res, data)


class GenericSyncClient(GenericClient):
    def send(self, req, stream=False):
        return self._client.send(req, stream=stream)

    def watch(self, br: BasicRequest, on_error: OnErrorHandler = on_error_raise):
        wd = WatchDriver(br, self._client.build_request, self._lazy)
        err_count = 0
        while True:
            req = wd.get_request(timeout=self._watch_timeout)
            resp = self.send(req, stream=True)
            try:
                resp.raise_for_status()
                err_count = 0
                for line in resp.iter_lines():
                    yield wd.process_one_line(line)
            except Exception as e:
                err_count += 1
                handle_error = on_error(e, err_count)
                if handle_error.action is OnErrorAction.RAISE:
                    raise
                if handle_error.action is OnErrorAction.STOP:
                    break
                if handle_error.sleep > 0:
                    time.sleep(handle_error.sleep)
                continue

    def request(
        self,
        method,
        res: Type[r.Resource] = None,
        obj=None,
        name=None,
        namespace=None,
        watch: bool = False,
        headers: dict = None,
        params: dict = None,
    ) -> Any:
        br = self.prepare_request(
            method, res, obj, name, namespace, watch, headers=headers, params=params
        )
        req = self.build_adapter_request(br)
        resp = self.send(req)
        return self.handle_response(method, resp, br)

    def list_chunks(self, br: BasicRequest) -> Iterator[Tuple[str, Iterator]]:
        cont = True
        while cont:
            req = self.build_adapter_request(br)
            resp = self.send(req)
            cont, rv, chunk = self.handle_response("list", resp, br)
            yield rv, chunk

    def list(self, br: BasicRequest) -> ListIterable:
        return ListIterable(self.list_chunks(br))


class GenericAsyncClient(GenericClient):
    AdapterClient = staticmethod(client_adapter.AsyncClient)

    async def send(self, req, stream=False):
        return await self._client.send(req, stream=stream)

    async def watch(self, br: BasicRequest, on_error: OnErrorHandler = on_error_raise):
        wd = WatchDriver(br, self._client.build_request, self._lazy)
        err_count = 0
        while True:
            req = wd.get_request(timeout=self._watch_timeout)
            resp = await self.send(req, stream=True)
            try:
                resp.raise_for_status()
                err_count = 0
                async for line in resp.aiter_lines():
                    yield wd.process_one_line(line)
            except Exception as e:
                err_count += 1
                handle_error = on_error(e, err_count)
                if handle_error.action is OnErrorAction.RAISE:
                    raise
                if handle_error.action is OnErrorAction.STOP:
                    break
                if handle_error.sleep > 0:
                    await asyncio.sleep(handle_error.sleep)
                continue
            finally:
                await resp.aclose()

    async def request(
        self,
        method,
        res: Type[r.Resource] = None,
        obj=None,
        name=None,
        namespace=None,
        watch: bool = False,
        headers: dict = None,
        params: dict = None,
    ) -> Any:
        br = self.prepare_request(
            method, res, obj, name, namespace, watch, headers=headers, params=params
        )
        req = self.build_adapter_request(br)
        resp = await self.send(req)
        return self.handle_response(method, resp, br)

    async def list_chunks(
        self, br: BasicRequest
    ) -> AsyncIterator[Tuple[str, Iterator]]:
        cont = True
        while cont:
            req = self.build_adapter_request(br)
            resp = await self.send(req)
            cont, rv, chunk = self.handle_response("list", resp, br)
            yield rv, chunk

    def list(self, br: BasicRequest) -> ListAsyncIterable:
        return ListAsyncIterable(self.list_chunks(br))

    async def close(self):
        await self._client.aclose()



import os
from pathlib import Path
from unittest.mock import patch

import pytest

from lightkube.config import kubeconfig
from lightkube.core import exceptions


def test_from_server():
    cfg = kubeconfig.KubeConfig.from_server("http://testserver.com").get()
    assert cfg.context_name == "default"
    assert cfg.namespace == "default"
    assert cfg.cluster.server == "http://testserver.com"
    assert cfg.user is None

    cfg = kubeconfig.KubeConfig.from_server("http://testserver.com", namespace="ns").get()
    assert cfg.context_name == "default"
    assert cfg.namespace == "ns"


@pytest.fixture()
def cfg():
    fname = Path("tests").joinpath("test_config.yaml")
    return kubeconfig.KubeConfig.from_file(fname)


def test_from_file(cfg):
    c = cfg.get()
    assert c.context_name == 'ctx11'
    assert c.user.username == 'u1'
    assert c.user.password == 'p1'
    assert c.cluster.server == 'server1'
    assert c.context.user == 'user1'
    assert c.namespace == kubeconfig.DEFAULT_NAMESPACE

    c = cfg.get(context_name='ctx12')
    assert c.context_name == 'ctx12'
    assert c.user.token == 'ABC'
    assert c.cluster.server == 'server1'
    assert c.context.user == 'user2'
    assert c.namespace == kubeconfig.DEFAULT_NAMESPACE

    c = cfg.get(context_name='ctx21')
    assert c.context_name == 'ctx21'
    assert c.user.username == 'u1'
    assert c.cluster.server == 'server2'
    assert c.context.cluster == 'cl2'
    assert c.namespace == 'ns21'


def test_from_file_miss_config(cfg):
    # non existing context raise an exception
    with pytest.raises(exceptions.ConfigError):
        assert cfg.get(context_name='ctx22')

    # if default context is missing, raise an exception
    cfg.current_context = None
    with pytest.raises(exceptions.ConfigError):
        assert cfg.get()

    # default context is missing, but a default is provided
    c = cfg.get(default=kubeconfig.PROXY_CONF)
    assert c is kubeconfig.PROXY_CONF


def test_from_dict():
    cfg = kubeconfig.KubeConfig.from_dict({
        'clusters': [{'name': 'cl1', 'cluster': {'server': 'a'}}],
        'contexts': [{'name': 'a', 'context': {'cluster': 'cl1', 'namespace': 'ns'}}]
    })
    assert cfg.current_context is None
    assert cfg.clusters['cl1'].server == 'a'
    assert cfg.contexts['a'].namespace == 'ns'

    c = cfg.get('a')
    assert c.namespace == 'ns'
    assert c.cluster.server == 'a'


@pytest.fixture
def service_account(tmpdir):
    tmpdir = Path(tmpdir)
    tmpdir.joinpath("namespace").write_text("my-namespace")
    tmpdir.joinpath("token").write_text("ABCD")
    tmpdir.joinpath("ca.crt").write_text("...bla...")

    os.environ["KUBERNETES_SERVICE_HOST"] = "k8s.local"
    os.environ["KUBERNETES_SERVICE_PORT"] = "9443"
    return tmpdir


def test_from_service_account(service_account):
    cfg = kubeconfig.KubeConfig.from_service_account(service_account)
    c = cfg.get()
    assert c.namespace == "my-namespace"
    assert c.user.token == "ABCD"
    assert c.cluster.server == "https://k8s.local:9443"


def test_from_service_account_not_found(tmpdir):
    with pytest.raises(exceptions.ConfigError):
        kubeconfig.KubeConfig.from_service_account(tmpdir)


def test_from_file_not_found(tmpdir):
    with pytest.raises(exceptions.ConfigError):
        kubeconfig.KubeConfig.from_file(Path(tmpdir).joinpath("bla"))


def test_from_env(service_account):
    cfg = kubeconfig.KubeConfig.from_env(service_account)
    assert cfg.get().user.token == "ABCD"

    with patch("lightkube.config.kubeconfig.os.environ") as environ:
        environ.get.return_value = str(Path("tests").joinpath("test_config.yaml"))
        cfg = kubeconfig.KubeConfig.from_env(service_account.joinpath("xyz"))
        assert cfg.get().context_name == 'ctx11'
        environ.get.assert_called_with("KUBECONFIG", kubeconfig.DEFAULT_KUBECONFIG)

    with patch("lightkube.config.kubeconfig.os.environ") as environ:
        environ.get.return_value = str(Path("tests").joinpath("test_config.yaml"))
        cfg = kubeconfig.KubeConfig.from_env(service_account.joinpath("xyz"), default_config='/tmp/bla')
        assert cfg.get().context_name == 'ctx11'
        environ.get.assert_called_with("KUBECONFIG", '/tmp/bla')



from typing import (
    Type,
    Union,
    overload,
    Dict,
    Tuple,
    List,
    Iterable,
    AsyncIterable,
)
import httpx

from ..config.kubeconfig import SingleConfig, KubeConfig
from ..core import resource as r
from .generic_client import GenericAsyncClient, ListAsyncIterable
from ..core.exceptions import ConditionError, ObjectDeleted
from ..types import OnErrorHandler, PatchType, CascadeType, on_error_raise
from .internal_resources import core_v1
from .selector import build_selector
from .client import (
    NamespacedResource,
    GlobalResource,
    GlobalSubResource,
    NamespacedSubResource,
    AllNamespacedResource,
    LabelSelector,
    FieldSelector,
)


class AsyncClient:
    """Creates a new lightkube client

    **parameters**

    * **config** - Instance of `SingleConfig` or `KubeConfig`. When not set the configuration will be detected automatically
      using the following order: in-cluster config, `KUBECONFIG` environment variable, `~/.kube/config` file.
    * **namespace** - Default namespace to use. This attribute is used in case namespaced resources are called without
      defining a namespace. If not specified, the default namespace set in your kube configuration will be used.
    * **timeout** - Instance of `httpx.Timeout`. By default all timeouts are set to 10 seconds. Notice that read timeout
      is ignored when watching changes.
    * **lazy** - When set, the returned objects will be decoded from the JSON payload in a lazy way, i.e. only when
      accessed.
    * **field_manager** - Name associated with the actor or entity that is making these changes.
    * **trust_env** - Ignore environment variables, also passed through to httpx.AsyncClient trust_env.  See its
      docs for further description. If False, empty config will be derived from_file(DEFAULT_KUBECONFIG)
    * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
        be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
        to `kubectl` commands.
    * **transport** - *(optional)* Custom httpx transport
    """

    def __init__(
        self,
        config: Union[SingleConfig, KubeConfig, None] = None,
        namespace: str = None,
        timeout: httpx.Timeout = None,
        lazy=True,
        field_manager: str = None,
        trust_env: bool = True,
        dry_run: bool = False,
        transport: httpx.AsyncBaseTransport = None,
    ):
        self._client = GenericAsyncClient(
            config,
            namespace=namespace,
            timeout=timeout,
            lazy=lazy,
            field_manager=field_manager,
            trust_env=trust_env,
            dry_run=dry_run,
            transport=transport,
        )

    @property
    def namespace(self):
        """Return the default namespace that will be used when a namespace has not been specified"""
        return self._client.namespace

    @property
    def config(self) -> SingleConfig:
        """Return the kubernetes configuration used in this client"""
        return self._client.config

    @overload
    async def delete(
        self,
        res: Type[GlobalResource],
        name: str,
        grace_period: int = None,
        cascade: CascadeType = None,
        dry_run: bool = False,
    ) -> None: ...

    @overload
    async def delete(
        self,
        res: Type[NamespacedResource],
        name: str,
        *,
        namespace: str = None,
        grace_period: int = None,
        cascade: CascadeType = None,
        dry_run: bool = False,
    ) -> None: ...

    async def delete(
        self,
        res,
        name: str,
        *,
        namespace: str = None,
        grace_period: int = None,
        cascade: CascadeType = None,
        dry_run: bool = False,
    ):
        """Delete an object

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to delete.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **grace_period** - *(optional)* The duration in seconds before the object should be deleted.
            Value must be non-negative integer. The value zero indicates delete immediately. If this value is `None`
            (default), the default grace period for the specified type will be used. Defaults to a per object value if
            not specified. Zero means delete immediately.
        * **cascade* - *(optional)* Whether and how garbage collection will be performed. Either this field or
            OrphanDependents may be set, but not both. The default policy is decided by the existing finalizer set
            in the metadata.finalizers and the resource-specific default policy. Acceptable values are:
            * 'CascadeType.ORPHAN' - orphan the dependents;
            * 'CascadeType.BACKGROUND' - allow the garbage collector to delete the dependents in the background;
            * 'CascadeType.FOREGROUND' - a cascading policy that deletes all dependents in the foreground.
        * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
            be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
            to `kubectl` commands.
        """
        return await self._client.request(
            "delete",
            res=res,
            name=name,
            namespace=namespace,
            params={
                "gracePeriodSeconds": grace_period,
                "propagationPolicy": cascade.value if cascade else None,
                "dryRun": "All" if dry_run else None,
            },
        )

    @overload
    async def deletecollection(
        self,
        res: Type[GlobalResource],
        grace_period: int = None,
        cascade: CascadeType = None,
        dry_run: bool = False,
    ) -> None: ...

    @overload
    async def deletecollection(
        self,
        res: Type[NamespacedResource],
        *,
        namespace: str = None,
        grace_period: int = None,
        cascade: CascadeType = None,
        dry_run: bool = False,
    ) -> None: ...

    async def deletecollection(
        self,
        res,
        *,
        namespace: str = None,
        grace_period: int = None,
        cascade: CascadeType = None,
        dry_run: bool = False,
    ):
        """Delete all objects of the given kind

        * **res** - Resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **grace_period** - *(optional)* The duration in seconds before the object should be deleted.
            Value must be non-negative integer. The value zero indicates delete immediately. If this value is `None`
            (default), the default grace period for the specified type will be used. Defaults to a per object value if
            not specified. Zero means delete immediately.
        * **cascade* - *(optional)* Whether and how garbage collection will be performed. Either this field or
            OrphanDependents may be set, but not both. The default policy is decided by the existing finalizer set
            in the metadata.finalizers and the resource-specific default policy. Acceptable values are:
            * 'CascadeType.ORPHAN' - orphan the dependents;
            * 'CascadeType.BACKGROUND' - allow the garbage collector to delete the dependents in the background;
            * 'CascadeType.FOREGROUND' - a cascading policy that deletes all dependents in the foreground.
        * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
            be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
            to `kubectl` commands.
        """
        return await self._client.request(
            "deletecollection",
            res=res,
            namespace=namespace,
            params={
                "gracePeriodSeconds": grace_period,
                "propagationPolicy": cascade.value if cascade else None,
                "dryRun": "All" if dry_run else None,
            },
        )

    @overload
    async def get(self, res: Type[GlobalResource], name: str) -> GlobalResource: ...

    @overload
    async def get(
        self, res: Type[AllNamespacedResource], name: str, *, namespace: str = None
    ) -> AllNamespacedResource: ...

    async def get(self, res, name, *, namespace=None):
        """Return an object

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to fetch.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return await self._client.request(
            "get", res=res, name=name, namespace=namespace
        )

    @overload
    def list(
        self,
        res: Type[GlobalResource],
        *,
        chunk_size: int = None,
        labels: LabelSelector = None,
        fields: FieldSelector = None,
    ) -> ListAsyncIterable[GlobalResource]: ...

    @overload
    def list(
        self,
        res: Type[NamespacedResource],
        *,
        namespace: str = None,
        chunk_size: int = None,
        labels: LabelSelector = None,
        fields: FieldSelector = None,
    ) -> ListAsyncIterable[NamespacedResource]: ...

    def list(self, res, *, namespace=None, chunk_size=None, labels=None, fields=None):
        """Return an iterator of objects matching the selection criteria.

        **parameters**

        * **res** - resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **chunk_size** - *(optional)* Limit the amount of objects returned for each rest API call.
             This method will automatically execute all subsequent calls until no more data is available.
        * **labels** - *(optional)* Limit the returned objects by labels. More [details](../selectors).
        * **fields** - *(optional)* Limit the returned objects by fields. More [details](../selectors).
        """

        br = self._client.prepare_request(
            "list",
            res=res,
            namespace=namespace,
            params={
                "limit": chunk_size,
                "labelSelector": build_selector(labels) if labels else None,
                "fieldSelector": (
                    build_selector(fields, for_fields=True) if fields else None
                ),
            },
        )
        return self._client.list(br)

    @overload
    def watch(
        self,
        res: Type[GlobalResource],
        *,
        labels: LabelSelector = None,
        fields: FieldSelector = None,
        server_timeout: int = None,
        resource_version: str = None,
        on_error: OnErrorHandler = on_error_raise,
    ) -> AsyncIterable[Tuple[str, GlobalResource]]: ...

    @overload
    def watch(
        self,
        res: Type[NamespacedResource],
        *,
        namespace: str = None,
        labels: LabelSelector = None,
        fields: FieldSelector = None,
        server_timeout: int = None,
        resource_version: str = None,
        on_error: OnErrorHandler = on_error_raise,
    ) -> AsyncIterable[Tuple[str, NamespacedResource]]: ...

    def watch(
        self,
        res,
        *,
        namespace=None,
        labels=None,
        fields=None,
        server_timeout=None,
        resource_version=None,
        on_error=on_error_raise,
    ):
        """Watch changes to objects

        **parameters**

        * **res** - resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **labels** - *(optional)* Limit the returned objects by labels. More [details](../selectors).
        * **fields** - *(optional)* Limit the returned objects by fields. More [details](../selectors).
        * **server_timeout** - *(optional)* Server side timeout in seconds to close a watch request.
            This method will automatically create a new request whenever the backend close the connection
            without errors.
        * **resource_version** - *(optional)* When set, only modification events following this version will be returned.
        * **on_error** - *(optional)* Function that control what to do in case of errors.
            The default implementation will raise any error.
        """
        br = self._client.prepare_request(
            "list",
            res=res,
            namespace=namespace,
            watch=True,
            params={
                "timeoutSeconds": server_timeout,
                "resourceVersion": resource_version,
                "labelSelector": build_selector(labels) if labels else None,
                "fieldSelector": (
                    build_selector(fields, for_fields=True) if fields else None
                ),
            },
        )
        return self._client.watch(br, on_error=on_error)

    @overload
    async def wait(
        self,
        res: Type[GlobalResource],
        name: str,
        *,
        for_conditions: Iterable[str],
        raise_for_conditions: Iterable[str] = (),
    ) -> GlobalResource: ...

    @overload
    async def wait(
        self,
        res: Type[AllNamespacedResource],
        name: str,
        *,
        for_conditions: Iterable[str],
        namespace: str = None,
        raise_for_conditions: Iterable[str] = (),
    ) -> AllNamespacedResource: ...

    async def wait(
        self,
        res,
        name: str,
        *,
        for_conditions: Iterable[str],
        namespace=None,
        raise_for_conditions: Iterable[str] = (),
    ):
        """Waits for specified conditions.

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of resource to wait for.
        * **for_conditions** - Condition types that are considered a success and will end the wait.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **raise_for_conditions** - *(optional)* Condition types that are considered failures and will exit the wait early.
        """

        kind = r.api_info(res).plural
        full_name = f"{kind}/{name}"

        for_conditions = list(for_conditions)
        raise_for_conditions = list(raise_for_conditions)

        watch = self.watch(res, namespace=namespace, fields={"metadata.name": name})
        try:
            async for op, obj in watch:

                if obj.status is None:
                    continue

                if op == "DELETED":
                    raise ObjectDeleted(full_name)

                try:
                    status = obj.status.to_dict()
                except AttributeError:
                    status = obj.status

                conditions = [
                    c for c in status.get("conditions", []) if c["status"] == "True"
                ]
                if any(c["type"] in for_conditions for c in conditions):
                    return obj

                failures = [c for c in conditions if c["type"] in raise_for_conditions]

                if failures:
                    raise ConditionError(
                        full_name, [f.get("message", f["type"]) for f in failures]
                    )
        finally:
            # we ensure the async generator is closed before returning
            await watch.aclose()

    @overload
    async def patch(
        self,
        res: Type[GlobalSubResource],
        name: str,
        obj: Union[GlobalSubResource, Dict, List],
        *,
        patch_type: PatchType = PatchType.STRATEGIC,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> GlobalSubResource: ...

    @overload
    async def patch(
        self,
        res: Type[GlobalResource],
        name: str,
        obj: Union[GlobalResource, Dict, List],
        *,
        patch_type: PatchType = PatchType.STRATEGIC,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> GlobalResource: ...

    @overload
    async def patch(
        self,
        res: Type[AllNamespacedResource],
        name: str,
        obj: Union[AllNamespacedResource, Dict, List],
        *,
        namespace: str = None,
        patch_type: PatchType = PatchType.STRATEGIC,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> AllNamespacedResource: ...

    async def patch(
        self,
        res,
        name,
        obj,
        *,
        namespace=None,
        patch_type=PatchType.STRATEGIC,
        field_manager=None,
        force=False,
        dry_run: bool = False,
    ):
        """Patch an object.

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to patch.
        * **obj** - patch object.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **patch_type** - *(optional)* Type of patch to execute. Default `PatchType.STRATEGIC`.
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
            **NOTE**: This parameter is mandatory (here or at `Client` creation time) for `PatchType.APPLY`.
        * **force** - *(optional)* Force is going to "force" Apply requests. It means user will re-acquire conflicting
          fields owned by other people. This parameter is ignored for non-apply patch types
        * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
            be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
            to `kubectl` commands.
        """
        force_param = "true" if force and patch_type == PatchType.APPLY else None
        return await self._client.request(
            "patch",
            res=res,
            name=name,
            namespace=namespace,
            obj=obj,
            headers={"Content-Type": patch_type.value},
            params={
                "force": force_param,
                "fieldManager": field_manager,
                "dryRun": "All" if dry_run else None,
            },
        )

    @overload
    async def create(
        self,
        obj: GlobalSubResource,
        name: str,
        field_manager: str = None,
        dry_run: bool = False,
    ) -> GlobalSubResource: ...

    @overload
    async def create(
        self,
        obj: NamespacedSubResource,
        name: str,
        *,
        namespace: str = None,
        field_manager: str = None,
        dry_run: bool = False,
    ) -> NamespacedSubResource: ...

    @overload
    async def create(
        self, obj: GlobalResource, field_manager: str = None, dry_run: bool = False
    ) -> GlobalResource: ...

    @overload
    async def create(
        self, obj: NamespacedResource, field_manager: str = None, dry_run: bool = False
    ) -> NamespacedResource: ...

    async def create(
        self,
        obj,
        name=None,
        *,
        namespace=None,
        field_manager=None,
        dry_run: bool = False,
    ):
        """Creates a new object

        **parameters**

        * **obj** - object to create. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
        * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
            be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
            to `kubectl` commands.
        """
        return await self._client.request(
            "post",
            name=name,
            namespace=namespace,
            obj=obj,
            params={
                "fieldManager": field_manager,
                "dryRun": "All" if dry_run else None,
            },
        )

    @overload
    async def replace(
        self,
        obj: GlobalSubResource,
        name: str,
        field_manager: str = None,
        dry_run: bool = False,
    ) -> GlobalSubResource: ...

    @overload
    async def replace(
        self,
        obj: NamespacedSubResource,
        name: str,
        *,
        namespace: str = None,
        field_manager: str = None,
        dry_run: bool = False,
    ) -> NamespacedSubResource: ...

    @overload
    async def replace(
        self, obj: GlobalResource, field_manager: str = None, dry_run: bool = False
    ) -> GlobalResource: ...

    @overload
    async def replace(
        self, obj: NamespacedResource, field_manager: str = None, dry_run: bool = False
    ) -> NamespacedResource: ...

    async def replace(
        self,
        obj,
        name=None,
        *,
        namespace=None,
        field_manager=None,
        dry_run: bool = False,
    ):
        """Replace an existing resource.

        **parameters**

        * **obj** - new object. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
        * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
            be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
            to `kubectl` commands.
        """
        return await self._client.request(
            "put",
            name=name,
            namespace=namespace,
            obj=obj,
            params={
                "fieldManager": field_manager,
                "dryRun": "All" if dry_run else None,
            },
        )

    @overload
    def log(
        self,
        name: str,
        *,
        namespace: str = None,
        container: str = None,
        follow: bool = False,
        since: int = None,
        tail_lines: int = None,
        timestamps: bool = False,
        newlines: bool = True,
    ) -> AsyncIterable[str]: ...

    def log(
        self,
        name,
        *,
        namespace=None,
        container=None,
        follow=False,
        since=None,
        tail_lines=None,
        timestamps=False,
        newlines=True,
    ):
        """Return log lines for the given Pod

        **parameters**

        * **name** - Name of the Pod.
        * **namespace** - *(optional)* Name of the namespace containing the Pod.
        * **container** - *(optional)* The container for which to stream logs. Defaults to only container if there is one container in the pod.
        * **follow** - *(optional)* If `True`, follow the log stream of the pod.
        * **since** - *(optional)* If set, a relative time in seconds before the current time from which to fetch logs.
        * **tail_lines** - *(optional)* If set, the number of lines from the end of the logs to fetch.
        * **timestamps** - *(optional)* If `True`, add an RFC3339 or RFC3339Nano timestamp at the beginning of every line of log output.
        * **newlines** - *(optional)* If `True`, each line will end with a newline, otherwise the newlines will be stripped.
        """
        br = self._client.prepare_request(
            "get",
            core_v1.PodLog,
            name=name,
            namespace=namespace,
            params={
                "timestamps": timestamps,
                "tailLines": tail_lines,
                "container": container,
                "sinceSeconds": since,
                "follow": follow,
            },
        )
        req = self._client.build_adapter_request(br)

        async def stream_log():
            resp = await self._client.send(req, stream=follow)
            self._client.raise_for_status(resp)
            async for line in resp.aiter_lines():
                yield line + "\n" if newlines else line

        return stream_log()

    @overload
    async def apply(
        self,
        obj: GlobalSubResource,
        name: str,
        *,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> GlobalSubResource: ...

    @overload
    async def apply(
        self,
        obj: NamespacedSubResource,
        name: str,
        *,
        namespace: str = None,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> NamespacedSubResource: ...

    @overload
    async def apply(
        self,
        obj: GlobalResource,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> GlobalResource: ...

    @overload
    async def apply(
        self,
        obj: NamespacedResource,
        field_manager: str = None,
        force: bool = False,
        dry_run: bool = False,
    ) -> NamespacedResource: ...

    async def apply(
        self,
        obj,
        name=None,
        *,
        namespace=None,
        field_manager=None,
        force=False,
        dry_run: bool = False,
    ):
        """Create or configure an object. This method uses the
        [server-side apply](https://kubernetes.io/docs/reference/using-api/server-side-apply/) functionality.

        **parameters**

        * **obj** - object to create. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - Name associated with the actor or entity that is making these changes.
        * **force** - *(optional)* Force is going to "force" Apply requests. It means user will re-acquire conflicting
          fields owned by other people.
        * **dry_run** - *(optional)* Apply server-side dry-run and guarantee that modifications will not
            be persisted in storage. Setting this field to `True` is equivalent of passing `--dry-run=server`
            to `kubectl` commands.
        """
        if (
            namespace is None
            and isinstance(obj, r.NamespacedResource)
            and obj.metadata.namespace
        ):
            namespace = obj.metadata.namespace
        if name is None and obj.metadata.name:
            name = obj.metadata.name
        return await self.patch(
            type(obj),
            name,
            obj,
            namespace=namespace,
            patch_type=PatchType.APPLY,
            field_manager=field_manager,
            force=force,
            dry_run=dry_run,
        )

    async def close(self):
        """Close the underline httpx client"""
        await self._client.close()



import base64
import shutil
import ssl
import unittest
from pathlib import Path
from unittest.mock import Mock
from lightkube.config import kubeconfig, client_adapter
from lightkube.config import models
from lightkube import ConfigError
import pytest
import httpx
import asyncio

BASEDIR = Path("tests")


def single_conf(cluster=None, user=None, fname=None):
    return kubeconfig.SingleConfig(
        context=models.Context(cluster="x"), context_name="x",
        cluster=cluster, user=user, fname=fname
    )


def test_verify_cluster_insecure():
    cfg = single_conf(cluster=models.Cluster(insecure=True), user=models.User())
    verify = client_adapter.verify_cluster(cfg.cluster, cfg.user, cfg.abs_file)
    assert verify.verify_mode is ssl.CERT_NONE
    assert not verify.check_hostname


def test_verify_cluster_secure():
    cfg = single_conf(cluster=models.Cluster(), user=models.User())
    verify = client_adapter.verify_cluster(cfg.cluster, cfg.user, cfg.abs_file)
    assert verify.verify_mode is ssl.CERT_REQUIRED


def get_issuer_mata(data: dict):
    return {d[0][0]: d[0][1] for d in data['issuer']}

def test_verify_cluster_ca_path(tmpdir):
    tmpdir = Path(tmpdir)
    data_dir = Path(__file__).parent.joinpath('data')
    shutil.copy(data_dir.joinpath("clientreq.pem"), tmpdir.joinpath("clientreq.pem"))
    cluster = models.Cluster(certificate_auth="clientreq.pem")
    cfg = single_conf(cluster=cluster, user=models.User(), fname=tmpdir.joinpath("kubeconf"))
    verify = client_adapter.verify_cluster(cfg.cluster, cfg.user, cfg.abs_file)
    assert get_issuer_mata(verify.get_ca_certs()[0])["organizationName"] == "Example"

    # fname not provided
    cfg = single_conf(cluster=models.Cluster(certificate_auth="clientreq.pem"), user=models.User())
    with pytest.raises(ConfigError):
        client_adapter.verify_cluster(cfg.cluster, cfg.user, cfg.abs_file)

    # cert path absolute
    cluster = models.Cluster(certificate_auth=str(data_dir.joinpath("clientreq.pem")))
    verify = client_adapter.verify_cluster(cluster, cfg.user, cfg.abs_file)
    assert get_issuer_mata(verify.get_ca_certs()[0])["organizationName"] == "Example"


def test_verify_cluster_ca_data():
    data_dir = Path(__file__).parent.joinpath('data')
    cert_data = base64.b64encode(data_dir.joinpath("clientreq.pem").read_bytes()).decode("utf8")

    cluster = models.Cluster(certificate_auth_data=cert_data)
    cfg = single_conf(cluster=cluster, user=models.User())
    verify = client_adapter.verify_cluster(cfg.cluster, cfg.user, cfg.abs_file)
    assert get_issuer_mata(verify.get_ca_certs()[0])["organizationName"] == "Example"


def test_user_cert_missing():
    cfg = single_conf(user=models.User())
    assert client_adapter.user_cert(cfg.user, cfg.abs_file) is None


def test_user_cert(tmpdir):
    tmpdir = Path(tmpdir)
    cfg = single_conf(user=models.User(client_cert="a.crt", client_key="a.key"), fname=tmpdir.joinpath("conf"))
    certs = client_adapter.user_cert(cfg.user, cfg.abs_file)
    assert certs == (tmpdir.joinpath("a.crt"), tmpdir.joinpath("a.key"))


def test_user_cert_data():
    cfg = single_conf(user=models.User(client_cert_data="Y2VydA==", client_key_data="a2V5"))
    certs = client_adapter.user_cert(cfg.user, cfg.abs_file)
    assert Path(certs[0]).read_text() == "cert"
    assert Path(certs[1]).read_text() == "key"


@unittest.mock.patch('ssl.create_default_context')
def test_verify_cluster_ca_and_cert(create_default_context):
    data_dir = Path(__file__).parent.joinpath('data')
    cluster = models.Cluster(certificate_auth=str(data_dir.joinpath("clientreq.pem")))
    cfg = single_conf(cluster=cluster, user=models.User(
        client_cert=str(data_dir.joinpath("clientreq.pem")),
        client_key=str(data_dir.joinpath("clientkey.pem"))
    ))
    verify = client_adapter.verify_cluster(cluster, cfg.user, cfg.abs_file)
    assert verify is create_default_context.return_value
    create_default_context.assert_called_once_with(cafile=str(data_dir.joinpath("clientreq.pem")))
    create_default_context.return_value.load_cert_chain.assert_called_once_with(
        str(data_dir.joinpath("clientreq.pem")),
        str(data_dir.joinpath("clientkey.pem"))
    )


def test_user_auth_missing():
    assert client_adapter.user_auth(None) is None


def test_user_auth_empty():
    assert client_adapter.user_auth(models.User()) is None


def test_user_auth_basic():
    auth = client_adapter.user_auth(models.User(username="user", password="psw"))
    assert isinstance(auth, httpx.BasicAuth)
    m = Mock(headers={})
    next(auth.auth_flow(m))
    assert m.headers["Authorization"] == "Basic dXNlcjpwc3c="


def test_user_auth_bearer():
    auth = client_adapter.user_auth(models.User(token="abcd"))
    assert isinstance(auth, client_adapter.BearerAuth)
    m = Mock(headers={})
    next(auth.auth_flow(m))
    assert m.headers["Authorization"] == "Bearer abcd"


def test_user_auth_provider():
    """Auth provider not supported"""
    with pytest.raises(ConfigError):
        client_adapter.user_auth(models.User(auth_provider={'x': 1}))


def test_user_auth_exec_sync():
    auth_script = str(Path(__file__).parent.joinpath('data', 'auth_script.sh'))
    auth = client_adapter.user_auth(models.User(exec=models.UserExec(
        apiVersion="client.authentication.k8s.io/v1beta1",
        command=auth_script,
    )))
    assert isinstance(auth, client_adapter.ExecAuth)
    m = Mock(headers={})
    next(auth.sync_auth_flow(m))
    assert m.headers["Authorization"] == "Bearer my-bearer-token"

    # call again should cache
    m = Mock(headers={})
    flow = auth.sync_auth_flow(m)
    next(flow)
    assert m.headers["Authorization"] == "Bearer my-bearer-token"
    m.headers["Authorization"] = None

    # we pretend the cache is old
    flow.send(httpx.Response(status_code=401, request=m))
    assert m.headers["Authorization"] == "Bearer my-bearer-token"


def test_user_auth_exec_sync_with_args():
    auth = client_adapter.user_auth(models.User(exec=models.UserExec(
        apiVersion="client.authentication.k8s.io/v1beta1",
        args=['{"apiVersion":"client.authentication.k8s.io/v1beta1",'
              '"kind":"ExecCredential","status":{"token":"my-bearer-token"}}'],
        command='echo',
    )))
    assert isinstance(auth, client_adapter.ExecAuth)
    m = Mock(headers={})
    next(auth.sync_auth_flow(m))
    assert m.headers["Authorization"] == "Bearer my-bearer-token"


def test_user_auth_exec_sync_fail():
    auth = client_adapter.user_auth(models.User(exec=models.UserExec(
        apiVersion="client.authentication.k8s.io/v1beta1",
        command="cp"
    )))
    with pytest.raises(ConfigError, match="cp"):
        next(auth.sync_auth_flow(Mock(headers={})))


@pytest.mark.asyncio
async def test_user_auth_exec_async():
    auth_script = str(Path(__file__).parent.joinpath('data', 'auth_script.sh'))
    auth = client_adapter.user_auth(models.User(exec=models.UserExec(
        apiVersion="client.authentication.k8s.io/v1beta1",
        command=auth_script,
    )))

    assert isinstance(auth, client_adapter.ExecAuth)
    m = Mock(headers={})
    await auth.async_auth_flow(m).__anext__()
    assert m.headers["Authorization"] == "Bearer my-bearer-token"

    # call again should cache
    m = Mock(headers={})
    flow = auth.async_auth_flow(m)
    await flow.__anext__()
    assert m.headers["Authorization"] == "Bearer my-bearer-token"
    m.headers["Authorization"] = None

    # we pretend the cache is old
    await flow.asend(httpx.Response(status_code=401, request=m))
    assert m.headers["Authorization"] == "Bearer my-bearer-token"
    with pytest.raises(StopAsyncIteration):
        await flow.__anext__()


@pytest.mark.asyncio
async def test_user_auth_exec_async_fail():
    auth = client_adapter.user_auth(models.User(exec=models.UserExec(
        apiVersion="client.authentication.k8s.io/v1beta1",
        command="cp"
    )))
    with pytest.raises(ConfigError, match="cp"):
        await auth.async_auth_flow(Mock(headers={})).__anext__()



from typing import Union, TextIO, Iterator, List, Mapping

import yaml

from .generic_resource import (
    GenericGlobalResource,
    GenericNamespacedResource,
    create_resources_from_crd,
)
from .core.exceptions import LoadResourceError
from .core.resource_registry import resource_registry

__all__ = ["from_dict", "load_all_yaml", "dump_all_yaml", "resource_registry"]

try:
    import jinja2
except ImportError:
    jinja2 = None

REQUIRED_ATTR = ("apiVersion", "kind")

AnyResource = Union[GenericGlobalResource, GenericNamespacedResource]


def from_dict(d: dict) -> AnyResource:
    """Converts a kubernetes resource defined as python dict to the corresponding resource object.
    If the dict represent a standard resource, the function will automatically load the appropriate
    resource type. Generic resources are also supported and used assuming they were defined prior to
    the function call. Returns the resource object or raise a `LoadResourceError`.

    **parameters**

    * **d** - A dictionary representing a Kubernetes resource. Keys `apiVersion` and `kind` are
      always required.
    """
    if not isinstance(d, Mapping):
        raise LoadResourceError("Invalid resource definition, not a dict.")
    for attr in REQUIRED_ATTR:
        if attr not in d:
            raise LoadResourceError(
                f"Invalid resource definition, key '{attr}' missing."
            )

    model = resource_registry.load(d["apiVersion"], d["kind"])
    return model.from_dict(d)


def load_all_yaml(
    stream: Union[str, TextIO],
    context: dict = None,
    template_env=None,
    create_resources_for_crds: bool = False,
) -> List[AnyResource]:
    """Load kubernetes resource objects defined as YAML. See `from_dict` regarding how resource types are detected.
    Returns a list of resource objects or raise a `LoadResourceError`.  Skips any empty YAML documents in the
    stream, returning an empty list if all YAML documents are empty. Deep parse any items from .*List resources.

    **parameters**

    * **stream** - A file-like object or a string representing a yaml file or a template resulting in
        a yaml file.
    * **context** - When is not `None` the stream is considered a `jinja2` template and the `context`
        will be used during templating.
    * **template_env** - `jinja2` template environment to be used for templating. When absent a standard
        environment is used.
    * **create_resources_for_crds** - If True, a generic resource will be created for every version
        of every CRD found that does not already have a generic resource.  There will be no side
        effect for any CRD that already has a generic resource.  Else if False, no generic resources
         will be created.  Default is False.

    **NOTE**: When using the template functionality (setting the context parameter), the dependency
        module `jinja2` need to be installed.
    """
    if context is not None:
        stream = _template(stream, context=context, template_env=template_env)

    def _flatten(objects: Iterator) -> List[AnyResource]:
        """Flatten resources which have a kind = *List."""
        resources = []
        for obj in objects:
            if obj is None:
                continue
            if isinstance(obj, Mapping) and obj.get("kind", "").endswith("List"):
                resources += _flatten(obj.get("items") or [])
            else:
                res = from_dict(obj)
                resources.append(res)

                if (
                    create_resources_for_crds is True
                    and res.kind == "CustomResourceDefinition"
                ):
                    create_resources_from_crd(res)
        return resources

    return _flatten(yaml.safe_load_all(stream))


def dump_all_yaml(resources: List[AnyResource], stream: TextIO = None, indent=2):
    """Write kubernetes resource objects as YAML into an open file.

    **parameters**

    * **resources** - List of resources to write on the file
    * **stream** - Path to a file where to write the resources. When not set the content is returned
      as a string.
    * **indent** - Number of characters for indenting nasted blocks.
    """
    res = [r.to_dict() for r in resources]
    return yaml.safe_dump_all(res, stream, indent=indent)


def _template(
    stream: Union[str, TextIO], context: dict = None, template_env=None
) -> List[AnyResource]:
    """
    Template a stream using jinja2 and the given context
    """
    if jinja2 is None:
        raise ImportError("load_from_template requires jinja2 to be installed")

    if template_env is None:
        template_env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
    elif not isinstance(template_env, jinja2.Environment):
        raise LoadResourceError("template_env is not a valid jinja2 template")

    tmpl = template_env.from_string(
        stream if isinstance(stream, str) else stream.read()
    )
    return tmpl.render(**context)



import json
import os
import ssl
import subprocess
from typing import Optional
import asyncio.subprocess

import httpx

from .kubeconfig import SingleConfig
from .models import Cluster, User, UserExec, FileStr
from ..core.exceptions import ConfigError


def Client(
    config: SingleConfig,
    timeout: httpx.Timeout,
    trust_env: bool = True,
    transport: httpx.BaseTransport = None,
) -> httpx.Client:
    return httpx.Client(
        transport=transport, **httpx_parameters(config, timeout, trust_env)
    )


def AsyncClient(
    config: SingleConfig,
    timeout: httpx.Timeout,
    trust_env: bool = True,
    transport: httpx.AsyncBaseTransport = None,
) -> httpx.AsyncClient:
    return httpx.AsyncClient(
        transport=transport, **httpx_parameters(config, timeout, trust_env)
    )


def httpx_parameters(config: SingleConfig, timeout: httpx.Timeout, trust_env: bool):
    return dict(
        timeout=timeout,
        base_url=config.cluster.server,
        verify=verify_cluster(config.cluster, config.user, config.abs_file, trust_env=trust_env),
        auth=user_auth(config.user),
        trust_env=trust_env,
    )


class BearerAuth(httpx.Auth):
    def __init__(self, token):
        self._bearer = f"Bearer {token}"

    def auth_flow(self, request: httpx.Request):
        request.headers["Authorization"] = self._bearer
        yield request


async def async_check_output(command, env):
    PIPE = asyncio.subprocess.PIPE
    proc = await asyncio.create_subprocess_exec(
        *command, env=env, stdin=None, stdout=PIPE, stderr=PIPE
    )
    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        raise ConfigError(
            f"Exec {command[0]} returned {proc.returncode}: {stderr.decode()}"
        )
    return stdout


def sync_check_output(command, env):
    proc = subprocess.Popen(
        command, env=env, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        raise ConfigError(
            f"Exec {command[0]} returned {proc.returncode}: {stderr.decode()}"
        )
    return stdout


class ExecAuth(httpx.Auth):
    def __init__(self, exec: UserExec):
        self._exec = exec
        self._last_bearer = None

    def _prepare(self):
        exec = self._exec
        if exec.apiVersion not in (
            "client.authentication.k8s.io/v1alpha1",
            "client.authentication.k8s.io/v1beta1",
        ):
            raise ConfigError(
                f"auth exec api version {exec.apiVersion} not implemented"
            )
        cmd_env_vars = dict(os.environ)
        if exec.env:
            cmd_env_vars.update((var.name, var.value) for var in exec.env)
        # TODO: add support for passing KUBERNETES_EXEC_INFO env var
        # https://github.com/kubernetes/community/blob/master/contributors/design-proposals/auth/kubectl-exec-plugins.md
        args = exec.args if exec.args else []
        return [exec.command] + args, cmd_env_vars

    def sync_auth_flow(self, request: httpx.Request):
        if self._last_bearer:
            request.headers["Authorization"] = self._last_bearer
            response = yield request
            if response.status_code != 401:
                return

        command, env = self._prepare()
        output = sync_check_output(command, env=env)
        token = json.loads(output)["status"]["token"]
        request.headers["Authorization"] = self._last_bearer = f"Bearer {token}"
        yield request

    async def async_auth_flow(self, request: httpx.Request):
        if self._last_bearer:
            request.headers["Authorization"] = self._last_bearer
            response = yield request
            if response.status_code != 401:
                return

        command, env = self._prepare()
        output = await async_check_output(command, env=env)
        token = json.loads(output)["status"]["token"]
        request.headers["Authorization"] = self._last_bearer = f"Bearer {token}"
        yield request


def user_auth(user: Optional[User]):
    if user is None:
        return None

    if user.token is not None:
        return BearerAuth(user.token)

    if user.exec:
        return ExecAuth(user.exec)

    if user.username and user.password:
        return httpx.BasicAuth(user.username, user.password)

    if user.auth_provider:
        raise ConfigError("auth-provider not supported")


def user_cert(user: User, abs_file):
    """Extract user certificates"""
    if user.client_cert or user.client_cert_data:
        return (
            FileStr(user.client_cert_data) or abs_file(user.client_cert),
            FileStr(user.client_key_data) or abs_file(user.client_key),
        )
    return None


def verify_cluster(cluster: Cluster, user: User, abs_file, trust_env: bool = True):
    """setup certificate verification"""
    if cluster.certificate_auth:
        ctx = ssl.create_default_context(cafile=abs_file(cluster.certificate_auth))
    elif cluster.certificate_auth_data:
        ctx = ssl.create_default_context(cafile=FileStr(cluster.certificate_auth_data))
    else:
        ctx = httpx.create_ssl_context(verify=not cluster.insecure, trust_env=trust_env)
    cert = user_cert(user, abs_file)
    if cert:
        ctx.load_cert_chain(*cert)
    return ctx



"""
Exceptions.
"""

import httpx

from .internal_models import meta_v1


class ConfigError(Exception):
    """
    Configuration specific errors.
    """

    pass


class NotReadyError(Exception):
    """
    Some information is not ready yet.
    """

    def __init__(self, name: str, message: str) -> None:
        super().__init__()
        self.name = name
        self.message = message

    def __str__(self) -> str:
        return f"{self.name} is not ready yet: {self.message}"


class ApiError(httpx.HTTPStatusError):
    status: "meta_v1.Status"

    def __init__(
        self, request: httpx.Request = None, response: httpx.Response = None
    ) -> None:
        self.status = meta_v1.Status.from_dict(response.json())
        super().__init__(self.status.message, request=request, response=response)


class LoadResourceError(Exception):
    """
    Error in loading a resource
    """


class ObjectDeleted(Exception):
    """
    Object was unexpectedly deleted
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name} was unexpectedly deleted"


class ConditionError(Exception):
    """
    Object is in specified bad condition
    """

    def __init__(self, name, messages):
        self.name = name
        self.messages = messages

    def __str__(self):
        messages = "; ".join(self.messages)
        return f"{self.name} has failure condition(s): {messages}"



import textwrap
from pathlib import Path
from unittest import mock

import pytest
import yaml

from lightkube import codecs
from lightkube.resources.core_v1 import ConfigMap
from lightkube.resources.rbac_authorization_v1 import Role
from lightkube.models.meta_v1 import ObjectMeta
from lightkube import generic_resource as gr
from lightkube import LoadResourceError
from lightkube.codecs import resource_registry

data_dir = Path(__file__).parent.joinpath('data')


@pytest.fixture(autouse=True)
def cleanup_registry():
    """Cleanup the registry before each test"""
    yield
    resource_registry.clear()


def test_from_dict():
    config_map = codecs.from_dict({
        'apiVersion': 'v1',
        'kind': 'ConfigMap',
        'metadata': {'name': 'config-name', 'labels': {'label1': 'value1'}},
        'data': {
            'file1.txt': 'some content here',
            'file2.txt': 'some other content'
        }
    })
    assert isinstance(config_map, ConfigMap)
    assert config_map.kind == 'ConfigMap'
    assert config_map.apiVersion == 'v1'
    assert config_map.metadata.name == 'config-name'
    assert config_map.metadata.labels['label1'] == 'value1'
    assert config_map.data['file1.txt'] == 'some content here'
    assert config_map.data['file2.txt'] == 'some other content'

    role = codecs.from_dict({
        'apiVersion': 'rbac.authorization.k8s.io/v1',
        'kind': 'Role',
        'metadata': {'name': 'read-pod'},
        'rules': [{
            'apiGroup': '',
            'resources': ['pods'],
            'verbs': ['get','watch', 'list']
        }]
    })
    assert isinstance(role, Role)
    assert role.kind == 'Role'
    assert role.apiVersion == 'rbac.authorization.k8s.io/v1'
    assert role.metadata.name == 'read-pod'
    assert role.rules[0].resources == ['pods']


def test_from_dict_wrong_model():
    # provided argument must actually be a dict
    with pytest.raises(LoadResourceError, match='.*not a dict'):
        codecs.from_dict([])

    # apiVersion and kind are required
    with pytest.raises(LoadResourceError, match=".*key 'apiVersion' missing"):
        codecs.from_dict({
            'kind': 'ConfigMap',
            'metadata': {'name': 'config-name'},
        })


def test_from_dict_generic_res():
    Mydb = gr.create_namespaced_resource('myapp.com', 'v1', 'Mydb', 'mydbs')
    db = codecs.from_dict({
        'apiVersion': 'myapp.com/v1',
        'kind': 'Mydb',
        'metadata': {'name': 'db1'},
        'key': {'a': 'b', 'c': 'd'}
    })
    assert isinstance(db, Mydb)
    assert db.kind == 'Mydb'
    assert db.apiVersion == 'myapp.com/v1'
    assert db.metadata.name == 'db1'
    assert 'key' in db
    assert db['key'] == {'a': 'b', 'c': 'd'}

    # Try with a generic resource with .k8s.io version
    # https://github.com/gtsystem/lightkube/issues/18
    version = "testing.k8s.io/v1"
    kind = "Testing"
    group, version_n = version.split("/")
    Testing = gr.create_namespaced_resource(group=group, version=version_n, kind=kind,
                                         plural=f"{kind.lower()}s")
    testing = codecs.from_dict({
        'apiVersion': version,
        'kind': kind,
        'metadata': {'name': 'testing1'},
        'key': {'a': 'b', 'c': 'd'}
    })
    assert isinstance(testing, Testing)
    assert testing.kind == kind
    assert testing.apiVersion == version
    assert testing.metadata.name == 'testing1'
    assert 'key' in testing
    assert testing['key'] == {'a': 'b', 'c': 'd'}


def test_from_dict_not_found():
    with pytest.raises(LoadResourceError):
        codecs.from_dict({'apiVersion': 'myapp2.com/v1', 'kind': 'Mydb'})

    with pytest.raises(LoadResourceError):
        codecs.from_dict({'apiVersion': 'v1', 'kind': 'Missing'})

    with pytest.raises(LoadResourceError):
        codecs.from_dict({'apiVersion': 'extra/v1', 'kind': 'Missing'})

    # Try with an undefined generic resource with .k8s.io version
    # https://github.com/gtsystem/lightkube/issues/18
    with pytest.raises(LoadResourceError):
        codecs.from_dict({'apiVersion': "undefined.k8s.io/v1", 'kind': 'Missing'})

@pytest.mark.parametrize(
    "yaml_file",
    (
            "example-def.yaml",
            "example-def-with-nulls.yaml",
            "example-def-with-lists.yaml"
    )
)
def test_load_all_yaml_static(yaml_file):
    gr.create_namespaced_resource('myapp.com', 'v1', 'Mydb', 'mydbs')
    objs = list(codecs.load_all_yaml(data_dir.joinpath(yaml_file).read_text()))
    kinds = [o.kind for o in objs]

    assert kinds == ['Secret', 'Mydb', 'Service', 'Deployment']

    with data_dir.joinpath('example-def.yaml').open() as f:
        objs = list(codecs.load_all_yaml(f))
    kinds = [o.kind for o in objs]

    assert kinds == ['Secret', 'Mydb', 'Service', 'Deployment']


def test_load_all_yaml_template():
    gr.create_namespaced_resource('myapp.com', 'v1', 'Mydb', 'mydbs')
    objs = list(codecs.load_all_yaml(
        data_dir.joinpath('example-def.tmpl').read_text(),
        context={'test': 'xyz'})
    )
    kinds = [o.kind for o in objs]

    assert kinds == ['Secret', 'Mydb', 'Service', 'Deployment']
    assert objs[1].metadata.name == 'bla-xyz'

    with data_dir.joinpath('example-def.tmpl').open() as f:
        objs = list(codecs.load_all_yaml(f, context={'test': 'xyz'}))
    kinds = [o.kind for o in objs]

    assert kinds == ['Secret', 'Mydb', 'Service', 'Deployment']
    assert objs[1].metadata.name == 'bla-xyz'


def test_load_all_yaml_template_env():
    gr.create_namespaced_resource('myapp.com', 'v1', 'Mydb', 'mydbs')
    import jinja2
    env = jinja2.Environment()
    env.globals['test'] = 'global'

    objs = list(codecs.load_all_yaml(
        data_dir.joinpath('example-def.tmpl').read_text(),
        context={},
        template_env=env)
    )
    kinds = [o.kind for o in objs]

    assert kinds == ['Secret', 'Mydb', 'Service', 'Deployment']
    assert objs[1].metadata.name == 'bla-global'

    with data_dir.joinpath('example-def.tmpl').open() as f:
        objs = list(codecs.load_all_yaml(f, context={}, template_env=env))
    kinds = [o.kind for o in objs]

    assert kinds == ['Secret', 'Mydb', 'Service', 'Deployment']
    assert objs[1].metadata.name == 'bla-global'

    # template_env is not an environment
    with pytest.raises(LoadResourceError, match='.*valid jinja2 template'):
        codecs.load_all_yaml(
            data_dir.joinpath('example-def.tmpl').read_text(),
            context={},
            template_env={}
        )


def test_load_all_yaml_all_null():
    yaml_file = "example-def-null.yaml"
    objs = list(codecs.load_all_yaml(data_dir.joinpath(yaml_file).read_text()))
    assert len(objs) == 0


@mock.patch('lightkube.codecs.jinja2', new=None)
def test_load_all_yaml_missing_dependency():
    with pytest.raises(ImportError, match='.*requires jinja2.*'):
        codecs.load_all_yaml(
            data_dir.joinpath('example-def.tmpl').read_text(),
            context={'test': 'xyz'}
        )


@pytest.mark.parametrize(
    "create_resources_for_crds",
    [
        True,   # generic resources should be created
        False,  # no generic resources should be created
    ]
)
def test_load_all_yaml_creating_generic_resources(create_resources_for_crds):
    template_yaml = data_dir.joinpath('example-multi-version-crd.yaml').read_text()
    template_dict = list(yaml.safe_load_all(template_yaml))[0]

    expected_group = template_dict["spec"]["group"]
    expected_kind = template_dict["spec"]["names"]["kind"]

    # Confirm no generic resources exist before testing
    assert len(resource_registry._registry) == 0

    objs = list(codecs.load_all_yaml(
        template_yaml,
        create_resources_for_crds=create_resources_for_crds,
    ))

    # Confirm expected resources exist
    if create_resources_for_crds:
        for version in template_dict["spec"]["versions"]:
            resource = resource_registry.get(f"{expected_group}/{version['name']}", expected_kind)
            assert resource is not None

        # Confirm we did not make any extra resources
        # + 1 as CustomResourceDefinition is also added to the registry
        assert len(resource_registry._registry) == len(template_dict["spec"]["versions"]) + 1
    else:
        # Confirm we did not make any resources except CustomResourceDefinition
        assert len(resource_registry._registry) == 1

    assert len(objs) == 1


def test_dump_all_yaml():
    cm = ConfigMap(
        apiVersion='v1', kind='ConfigMap',
        metadata=ObjectMeta(name='xyz', labels={'x': 'y'})
    )
    Mydb = gr.create_namespaced_resource('myapp.com', 'v1', 'Mydb', 'mydbs')

    db = Mydb(
        apiVersion='myapp.com/v1', kind='Mydb',
        metadata=ObjectMeta(name='db1'), xyz={'a': 'b'}
    )

    res = codecs.dump_all_yaml([cm, db])
    expected = textwrap.dedent("""
        apiVersion: v1
        kind: ConfigMap
        metadata:
          labels:
            x: y
          name: xyz
        ---
        apiVersion: myapp.com/v1
        kind: Mydb
        metadata:
          name: db1
        xyz:
          a: b
    """).lstrip()
    assert res == expected

    res = codecs.dump_all_yaml([db, cm], indent=4)
    expected = textwrap.dedent("""
        apiVersion: myapp.com/v1
        kind: Mydb
        metadata:
            name: db1
        xyz:
            a: b
        ---
        apiVersion: v1
        kind: ConfigMap
        metadata:
            labels:
                x: y
            name: xyz
    """).lstrip()
    assert res == expected




from typing import List
from dataclasses import dataclass, field
from datetime import datetime, timezone

import pytest

from lightkube.core.dataclasses_dict import DataclassDictMixIn


@dataclass
class B(DataclassDictMixIn):
    b1: str
    b2: 'A' = None
    b3: 'dict' = None


@dataclass
class C(DataclassDictMixIn):
    c1: str
    c2: List['A'] = None
    c3: str = field(metadata={"json": "$ref"}, default=None)


@dataclass
class A(DataclassDictMixIn):
    a1: str
    a2: int = 0
    a3: 'bool' = False


@dataclass
class DT(DataclassDictMixIn):
    dt: 'datetime'


@dataclass
class Def(DataclassDictMixIn):
    d1: str
    d2: int = 2
    d3: 'bool' = False
    d4: str = "ok"


@pytest.mark.parametrize("lazy", [True, False])
def test_lazy_default_not_hiding_added(lazy):
    # Setup a C object without setting c2 (default)
    inst = C.from_dict({"c1": "val"}, lazy=lazy)
    assert inst.to_dict() == {"c1": "val"}
    inst.c2 = [A("def")]
    assert inst.to_dict() == {"c1": "val", "c2": [{"a1": "def"}]}


@pytest.mark.parametrize("lazy", [True, False])
def test_lazy_loaded_not_hiding_set(lazy):
    # Setup a C object without setting c2 (default)
    inst = C.from_dict({"c1": "val", "c2": [{"a1": "abc"}]}, lazy=lazy)
    assert inst.to_dict() == {"c1": "val", "c2": [{"a1": "abc"}]}
    # Change c2 list attribute
    inst.c2 = [A("def")]
    assert inst.to_dict() == {"c1": "val", "c2": [{"a1": "def"}]}


@pytest.mark.parametrize("lazy", [True, False])
def test_lazy_loaded_not_hiding_cleared(lazy):
    # Setup a C object without setting c2 (default)
    inst = C.from_dict({"c1": "val", "c2": [{"a1": "abc"}]}, lazy=lazy)
    assert inst.to_dict() == {"c1": "val", "c2": [{"a1": "abc"}]}
    # Change c2 list attribute
    inst.c2 = None
    assert inst.to_dict() == {"c1": "val"}


@pytest.mark.parametrize("lazy", [True, False])
def test_single(lazy):
    a = A.from_dict({'a1': 'a', 'a3': True}, lazy=lazy)
    assert a.a1 == 'a'
    assert a.a2 == 0
    assert a.a3 is True
    assert a.to_dict() == {'a1': 'a', 'a3': True}


@pytest.mark.parametrize("lazy", [True, False])
def test_nasted(lazy):
    b = B.from_dict({'b1': 'ok', 'b2': {'a1': 'a', 'a3': True}}, lazy=lazy)
    assert b.b1 == 'ok'
    if lazy:    # when we use lazy, sub-objects are not expanded yet
        assert 'b2' not in vars(b)
    else:
        assert 'b2' in vars(b)
    assert b.b2.a3 is True
    assert b.to_dict() == {'b1': 'ok', 'b2': {'a1': 'a', 'a3': True}}


@pytest.mark.parametrize("lazy", [True, False])
def test_nasted_in_list(lazy):
    c = C.from_dict({'c1': 'ok', 'c2': [{'a1': 'a', 'a3': True}, {'a1': 'b'}]}, lazy=lazy)
    if lazy:  # when we use lazy, sub-objects are not expanded yet
        assert 'c2' not in vars(c)
    else:
        assert 'c2' in vars(c)
    assert c.c2[0].a3 is True
    assert c.c2[1].a1 == 'b'
    assert c.to_dict() == {'c1': 'ok', 'c2': [
        {'a1': 'a', 'a3': True}, {'a1': 'b'}
    ]}


def test_nasted_to_dict():
    b = B.from_dict({'b1': 'ok', 'b2': {'a1': 'a', 'a3': True}}, lazy=True)
    assert b.to_dict() == {'b1': 'ok', 'b2': {'a1': 'a', 'a3': True}}


@pytest.mark.parametrize("lazy", [True, False])
def test_dict(lazy):
    b = B.from_dict({'b1': 'ok', 'b3': {'xx': 'x'}}, lazy=lazy)
    assert b.to_dict() == {'b1': 'ok', 'b3': {'xx': 'x'}}


def test_datatime():
    """Datetime get converted to string and back"""
    d = DT.from_dict({'dt': '2019-08-03T11:32:48Z'})
    assert d.dt == datetime(2019, 8, 3, 11, 32, 48, tzinfo=timezone.utc)
    assert d.to_dict() == {'dt': '2019-08-03T11:32:48Z'}

    d = DT.from_dict({'dt': '2019-08-03T11:32:48+02:30'})
    assert isinstance(d.dt, datetime) and str(d.dt) == '2019-08-03 11:32:48+02:30'
    assert d.to_dict() == {'dt': '2019-08-03T11:32:48+02:30'}


@pytest.mark.parametrize("lazy", [True, False])
def test_rename(lazy):
    """We can rename fields from/to dicts"""
    c = C.from_dict({'c1': 'a', '$ref': 'b'}, lazy=lazy)
    assert c.c1 == 'a'
    assert c.c3 == 'b'
    c.c3 = 'c'

    assert c.to_dict() == {'c1': 'a', '$ref': 'c'}


@pytest.mark.parametrize("lazy", [True, False])
def test_drop_unknown(lazy):
    """Unknown attributes are dropped"""
    c = C.from_dict({'c1': 'a', 'k': 'b'}, lazy=lazy)
    assert c.c1 == 'a'
    assert not hasattr(c, 'k')

    assert c.to_dict() == {'c1': 'a'}


def test_default_not_encoded():
    """Test that default values are not returned in the dict"""
    assert Def(d1='a').to_dict() == {'d1': 'a'}
    assert Def(d1='a', d2=2).to_dict() == {'d1': 'a'}
    assert Def(d1='a', d2=0).to_dict() == {'d1': 'a', 'd2': 0}
    assert Def(d1='a', d3=False).to_dict() == {'d1': 'a'}
    assert Def(d1='a', d3=True).to_dict() == {'d1': 'a', 'd3': True}
    assert Def(d1='a', d4='ok').to_dict() == {'d1': 'a'}
    assert Def(d1='a', d4='ko').to_dict() == {'d1': 'a', 'd4': 'ko'}



import enum
from dataclasses import dataclass
import typing


class PatchType(enum.Enum):
    JSON = "application/json-patch+json"
    MERGE = "application/merge-patch+json"
    STRATEGIC = "application/strategic-merge-patch+json"
    APPLY = "application/apply-patch+yaml"


class CascadeType(enum.Enum):
    ORPHAN = "Orphan"
    BACKGROUND = "Background"
    FOREGROUND = "Foreground"


class OnErrorAction(enum.Enum):
    RETRY = 0  # retry to perform the API call again from the last version
    STOP = 1  # stop silently the iterator
    RAISE = 2  # raise the error on the caller scope


@dataclass
class OnErrorResult:
    action: OnErrorAction
    sleep: float = 0


OnErrorHandler = typing.Callable[[Exception, int], OnErrorResult]


def on_error_raise(e, count):
    return OnErrorResult(OnErrorAction.RAISE)


def on_error_stop(e, count):
    return OnErrorResult(OnErrorAction.STOP)


def on_error_retry(e, count):
    return OnErrorResult(OnErrorAction.RETRY)



import pytest

from lightkube.codecs import resource_registry
from lightkube.resources.core_v1 import Pod
from lightkube.resources.apps_v1 import Deployment
from lightkube.resources.events_v1 import Event
from lightkube.core import resource as res

@pytest.fixture(autouse=True)
def cleanup_registry():
    """Cleanup the registry after each test"""
    yield
    resource_registry.clear()

@pytest.mark.parametrize(
    "version,kind,Res",
    [("v1", "Pod", Pod), ("apps/v1", "Deployment", Deployment), ("events.k8s.io/v1", "Event", Event)]
)
def test_register(version, kind, Res):
    assert resource_registry.get(version, kind) is None
    res = resource_registry.register(Res)
    assert res is Res
    assert resource_registry.get(version, kind) is Res


def test_register_decorator():
    assert resource_registry.get("test.io/v1", "Search") is None

    @resource_registry.register
    class Search(res.NamespacedResource):
        _api_info = res.ApiInfo(
            resource=res.ResourceDef('test.io', 'v1', 'Search'),
            plural='searches',
            verbs=['delete', 'deletecollection', 'get', 'global_list', 'global_watch', 'list', 'patch', 'post', 'put',
                   'watch']
        )

    assert resource_registry.get("test.io/v1", "Search") is Search


@pytest.mark.parametrize(
    "version,kind,Res",
    [("v1", "Pod", Pod), ("apps/v1", "Deployment", Deployment), ("events.k8s.io/v1", "Event", Event)]
)
def test_load(version, kind, Res):
    assert resource_registry.get(version, kind) is None
    pod_class = resource_registry.load(version, kind)
    assert pod_class is Res
    assert resource_registry.get(version, kind) is Res



import decimal
import re
from typing import Optional, overload

from ..core.internal_models import core_v1

ResourceRequirements = core_v1.ResourceRequirements

MULTIPLIERS = {
    # Bytes
    "m": (10, -3),  # 1000^(-1) (=0.001)
    "": (10, 0),  # 1000^0 (=1)
    "k": (10, 3),  # 1000^1
    "M": (10, 6),  # 1000^2
    "G": (10, 9),  # 1000^3
    "T": (10, 12),  # 1000^4
    "P": (10, 15),  # 1000^5
    "E": (10, 18),  # 1000^6
    "Z": (10, 21),  # 1000^7
    "Y": (10, 24),  # 1000^8
    # Bibytes
    "Ki": (1024, 1),  # 2^10
    "Mi": (1024, 2),  # 2^20
    "Gi": (1024, 3),  # 2^30
    "Ti": (1024, 4),  # 2^40
    "Pi": (1024, 5),  # 2^50
    "Ei": (1024, 6),  # 2^60
    "Zi": (1024, 7),  # 2^70
    "Yi": (1024, 8),  # 2^80
}

# Pre-calculate multipliers and store as decimals.
MULTIPLIERS = {k: decimal.Decimal(v[0]) ** v[1] for k, v in MULTIPLIERS.items()}


def parse_quantity(quantity: Optional[str]) -> Optional[decimal.Decimal]:
    """Parse a quantity string into a bare (suffix-less) decimal.

    Kubernetes converts user input to a canonical representation. For example, "0.9Gi" would be converted
    to "966367641600m".
    This function can be useful for comparing user input to actual values, for example comparing
    resource limits between a StatefulSet's template
    (`statefulset.spec.template.spec.containers[i].resources`) and a scheduled pod
    (`pod.spec.containers[i].resources`) after patching the StatefulSet.

    **Parameters**

    * **quantity** `str` - An str representing a K8s quantity (e.g. "1Gi" or "1G"), per
    https://kubernetes.io/docs/reference/kubernetes-api/common-definitions/quantity/.

    **returns**  An instance of `decimal.Decimal` representing the quantity as a bare decimal.
    """
    if quantity is None:
        # This is useful for comparing e.g. ResourceRequirements.limits.get("cpu"), which can be
        # None.
        return None

    pat = re.compile(r"([+-]?\d+(?:[.]\d*)?(?:e[+-]?\d+)?|[.]\d+(?:e[+-]?\d+)?)(.*)")
    match = pat.match(quantity)

    if not match:
        raise ValueError("Invalid quantity string: '{}'".format(quantity))

    try:
        value = decimal.Decimal(match.group(1))
    except ArithmeticError as e:
        raise ValueError("Invalid numerical value") from e

    unit = match.group(2)

    try:
        multiplier = MULTIPLIERS[unit]
    except KeyError:
        raise ValueError("Invalid unit suffix: {}".format(unit))

    try:
        as_decimal = value * multiplier
        return as_decimal.quantize(decimal.Decimal("0.001"), rounding=decimal.ROUND_UP)
    except ArithmeticError as e:
        raise ValueError("Invalid numerical value") from e


def _equals_canonically(
    first_dict: Optional[dict], second_dict: Optional[dict]
) -> bool:
    """Compare resource dicts such as 'limits' or 'requests'."""
    if first_dict == second_dict:
        # This covers two cases: (1) both args are None; (2) both args are identical dicts.
        return True
    if first_dict and second_dict:
        if first_dict.keys() != second_dict.keys():
            # The dicts have different keys, so they cannot possibly be equal
            return False
        return all(
            parse_quantity(first_dict[k]) == parse_quantity(second_dict[k])
            for k in first_dict.keys()
        )
    if not first_dict and not second_dict:
        # This covers cases such as first=None and second={}
        return True
    return False


@overload
def equals_canonically(
    first: ResourceRequirements, second: ResourceRequirements
) -> bool: ...


@overload
def equals_canonically(first: Optional[dict], second: Optional[dict]) -> bool: ...


def equals_canonically(first, second):
    """Compare two resource requirements for numerical equality.

    Both arguments must be of the same type and can be either:

    - An instance of `core_v1.ResourceRequirements`
    - `Optional[dict]`: representing the "limits" or the "requests" portion of `ResourceRequirements`.

    ```python
    >>> equals_canonically({"cpu": "0.6"}, {"cpu": "600m"})
    True

    >>> equals_canonically(
            ResourceRequirements(limits={"cpu": "0.6"}),
            ResourceRequirements(limits={"cpu": "600m"})
        )
    True
    ```

    **Parameters**

    * **first** `ResourceRequirements` or `dict` - The first item to compare.
    * **second** `ResourceRequirements` or `dict` - The second item to compare.

    **returns**  True, if both arguments are numerically equal; False otherwise.
    """
    if isinstance(first, (dict, type(None))) and isinstance(second, (dict, type(None))):
        # Args are 'limits' or 'requests' dicts
        return _equals_canonically(first, second)
    elif isinstance(first, ResourceRequirements) and isinstance(
        second, ResourceRequirements
    ):
        # Args are ResourceRequirements, which may contain 'limits' and 'requests' dicts
        ks = ("limits", "requests")
        return all(
            _equals_canonically(getattr(first, k), getattr(second, k)) for k in ks
        )
    else:
        raise TypeError(
            "unsupported operand type(s) for canonical comparison: '{}' and '{}'".format(
                first.__class__.__name__,
                second.__class__.__name__,
            )
        )



import sys

try:
    from ..resources import core_v1

    try:
        from ..resources import apiextensions_v1 as apiextensions
    except:
        from ..resources import apiextensions_v1beta1 as apiextensions
except:
    if (
        sys.modules["__main__"].__package__ != "mkdocs"
    ):  # we ignore this import error during documentation generation
        raise
    from unittest import mock

    class CustomResourceDefinition:
        pass

    apiextensions = mock.Mock()
    apiextensions.CustomResourceDefinition = CustomResourceDefinition

    class PodLog:
        pass

    core_v1 = mock.Mock()
    core_v1.PodLog = PodLog



from .core.client import Client
from .core.async_client import AsyncClient
from .core.generic_client import ALL_NS
from .core.exceptions import ApiError, NotReadyError, ConfigError, LoadResourceError
from .core.sort_objects import sort_objects
from .config.kubeconfig import KubeConfig, SingleConfig

__all__ = [
    "Client",
    "AsyncClient",
    "ALL_NS",
    "ApiError",
    "NotReadyError",
    "ConfigError",
    "LoadResourceError",
    "sort_objects",
    "KubeConfig",
    "SingleConfig",
]



import sys

try:
    from ..models import meta_v1, autoscaling_v1, core_v1

except:
    if (
        sys.modules["__main__"].__package__ != "mkdocs"
    ):  # we ignore this import error during documentation generation
        raise
    from unittest import mock

    class ObjectMeta:
        pass

    meta_v1 = mock.Mock()
    meta_v1.ObjectMeta = ObjectMeta

    class Scale:
        pass

    autoscaling_v1 = mock.Mock()
    autoscaling_v1.Scale = Scale

    class ResourceRequirements:
        pass

    core_v1 = mock.Mock()
    core_v1.ResourceRequirements = ResourceRequirements



from collections import namedtuple
import unittest.mock
import warnings
from ssl import SSLContext

warnings.filterwarnings("ignore", category=DeprecationWarning)

import json
import pytest
import httpx
import respx

import lightkube
from lightkube.config.kubeconfig import KubeConfig, SingleConfig, Context, Cluster, User
from lightkube.resources.core_v1 import Pod, Node, Binding
from lightkube.models.meta_v1 import ObjectMeta
from lightkube import types
from lightkube.generic_resource import create_global_resource

KUBECONFIG = """
apiVersion: v1
clusters:
- cluster: {server: 'https://localhost:9443'}
  name: test
contexts:
- context: {cluster: test, user: test}
  name: test
current-context: test
kind: Config
preferences: {}
users:
- name: test
  user: {token: testtoken}
"""


def json_contains(json_str, data: dict):
    obj = json.loads(json_str)
    for key, value in data.items():
        assert key in obj
        assert obj[key] == value


@pytest.fixture
def kubeconfig(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(KUBECONFIG)
    return kubeconfig


@pytest.fixture
def kubeconfig_ns(tmpdir):
    kubeconfig = tmpdir.join("kubeconfig")
    kubeconfig.write(KUBECONFIG.replace('user: test', 'user: test, namespace: ns1'))
    return kubeconfig


@pytest.fixture
def client(kubeconfig):
    config = KubeConfig.from_file(str(kubeconfig))
    return lightkube.Client(config=config)


def test_namespace(client: lightkube.Client, kubeconfig_ns):
    assert client.namespace == 'default'

    config = KubeConfig.from_file(str(kubeconfig_ns))
    client = lightkube.Client(config=config)
    assert client.namespace == 'ns1'


def test_client_config_attribute(kubeconfig):
    config = KubeConfig.from_file(kubeconfig)
    client = lightkube.Client(config=config)
    assert client.config == config.get()

    single_conf = config.get()
    client = lightkube.Client(config=single_conf)
    assert client.config is single_conf


@unittest.mock.patch('lightkube.core.generic_client.KubeConfig')
def test_client_default_config_construction(mock_kube_config):
    config = SingleConfig(
        context_name="test",
        context=Context(cluster='test', user="test"),
        cluster=Cluster(server="https://localhost:9443"),
        user=User(username="test"),
    )

    # config should be generated from environment without trust_env
    mock_kube_config.from_env.return_value.get.return_value = config
    lightkube.Client()
    mock_kube_config.from_env.assert_called_once_with()
    mock_kube_config.from_file.assert_not_called()

    # trust_env=False should create from DEFAULT_KUBECONFIG
    mock_kube_config.reset_mock()
    mock_kube_config.from_file.return_value.get.return_value = config
    lightkube.Client(trust_env=False)
    mock_kube_config.from_file.assert_called_once_with('~/.kube/config')
    mock_kube_config.from_env.assert_not_called()


@unittest.mock.patch('httpx.Client')
@unittest.mock.patch('lightkube.config.client_adapter.user_auth')
@unittest.mock.patch('lightkube.config.client_adapter.verify_cluster')
def test_client_httpx_attributes(verify_cluster, user_auth, httpx_client, kubeconfig):
    config = KubeConfig.from_file(kubeconfig)
    single_conf = config.get()
    lightkube.Client(config=single_conf, trust_env=False)
    verify_cluster.assert_called_once_with(single_conf.cluster, single_conf.user, single_conf.abs_file, trust_env=False)
    httpx_client.assert_called_once_with(
        timeout=None,
        base_url=single_conf.cluster.server,
        verify=verify_cluster.return_value,
        auth=user_auth.return_value,
        trust_env=False,
        transport=None,
    )


@respx.mock
def test_get_namespaced(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = client.get(Pod, name="xx")
    assert pod.metadata.name == 'xx'

    respx.get("https://localhost:9443/api/v1/namespaces/other/pods/xx").respond(json={'metadata': {'name': 'xy'}})
    pod = client.get(Pod, name="xx", namespace="other")
    assert pod.metadata.name == 'xy'


@respx.mock
def test_get_global(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/nodes/n1").respond(json={'metadata': {'name': 'n1'}})
    pod = client.get(Node, name="n1")
    assert pod.metadata.name == 'n1'

    # GET doesn't support all namespaces
    with pytest.raises(ValueError):
        client.get(Pod, name="xx", namespace=lightkube.ALL_NS)


@respx.mock
def test_list_namespaced(client: lightkube.Client):
    resp = {'items':[{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}],
            'metadata': {'resourceVersion': '42'}}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods").respond(json=resp)
    pods = client.list(Pod)
    with pytest.raises(lightkube.NotReadyError):
        pods.resourceVersion
    for pod, expected in zip(pods, resp["items"]):
        assert pod.metadata is not None
        assert pod.metadata.name == expected["metadata"]["name"]
        assert pod.apiVersion is not None
        assert pod.kind is not None
    assert pods.resourceVersion == "42"

    respx.get("https://localhost:9443/api/v1/namespaces/other/pods?labelSelector=k%3Dv").respond(json=resp)
    pods = client.list(Pod, namespace="other", labels={'k': 'v'})
    assert [pod.metadata.name for pod in pods] == ['xx', 'yy']


@respx.mock
def test_list_crd(client: lightkube.Client):
    """CRD list seems to return always the 'continue' metadata attribute"""
    resp = {'items': [{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}], 'metadata': {'continue': ''}}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods").respond(json=resp)
    pods = client.list(Pod)
    assert [pod.metadata.name for pod in pods] == ['xx', 'yy']


@respx.mock
def test_list_global(client: lightkube.Client):
    resp = {'items': [{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}]}
    respx.get("https://localhost:9443/api/v1/nodes").respond(json=resp)
    nodes = client.list(Node)
    assert [node.metadata.name for node in nodes] == ['xx', 'yy']

    respx.get("https://localhost:9443/api/v1/pods?fieldSelector=k%3Dx").respond(json=resp)
    pods = client.list(Pod, namespace=lightkube.ALL_NS, fields={'k': 'x'})
    assert [pod.metadata.name for pod in pods] == ['xx', 'yy']

    # Binding doesn't support all namespaces
    with pytest.raises(ValueError):
        client.list(Binding, namespace=lightkube.ALL_NS)


@respx.mock
def test_list_chunk_size(client: lightkube.Client):
    resp = {'items': [{'metadata': {'name': 'xx'}}, {'metadata': {'name': 'yy'}}], 'metadata': {'continue': 'yes'}}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods?limit=3").respond(json=resp)
    resp = {'items': [{'metadata': {'name': 'zz'}}]}
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods?limit=3&continue=yes").respond(json=resp)
    pods = client.list(Pod, chunk_size=3)
    assert [pod.metadata.name for pod in pods] == ['xx', 'yy', 'zz']


@respx.mock
def test_delete_namespaced(client: lightkube.Client):
    respx.delete("https://localhost:9443/api/v1/namespaces/default/pods/xx")
    client.delete(Pod, name="xx")

    respx.delete("https://localhost:9443/api/v1/namespaces/other/pods/xx")
    client.delete(Pod, name="xx", namespace="other")

    # test grace_period parameter
    respx.delete("https://localhost:9443/api/v1/namespaces/default/pods/x_grace?gracePeriodSeconds=30")
    client.delete(Pod, name="x_grace", grace_period=30)

    # test cascade parameter
    respx.delete("https://localhost:9443/api/v1/namespaces/default/pods/x_cascade?propagationPolicy=Background")
    client.delete(Pod, name="x_cascade", cascade=types.CascadeType.BACKGROUND)

    # test dry_run parameter
    req_dry = respx.delete("https://localhost:9443/api/v1/namespaces/other/pods/z?dryRun=All")
    pod = client.delete(Pod, name="z", namespace="other", dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'


@respx.mock
def test_delete_global(client: lightkube.Client):
    respx.delete("https://localhost:9443/api/v1/nodes/xx")
    client.delete(Node, name="xx")

    # test dry_run parameter
    req_dry = respx.delete("https://localhost:9443/api/v1/nodes/z?dryRun=All")
    node = client.delete(Node, name="z", dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'

@respx.mock
def test_delete_collection_namespaced(client: lightkube.Client):

    # test dry_run parameter
    req_dry = respx.delete("https://localhost:9443/api/v1/namespaces/other/pods?dryRun=All")
    pod = client.deletecollection(Pod, namespace="other", dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'

    respx.delete("https://localhost:9443/api/v1/namespaces/default/pods")
    client.deletecollection(Pod)

    respx.delete("https://localhost:9443/api/v1/namespaces/other/pods")
    client.deletecollection(Pod, namespace="other")

    # test grace_period parameter
    respx.delete("https://localhost:9443/api/v1/namespaces/grace/pods?gracePeriodSeconds=30")
    client.deletecollection(Pod, namespace="grace", grace_period=30)

    # test cascade parameter
    respx.delete("https://localhost:9443/api/v1/namespaces/cascade/pods?propagationPolicy=Orphan")
    client.deletecollection(Pod, namespace="cascade", cascade=types.CascadeType.ORPHAN)

@respx.mock
def test_deletecollection_global(client: lightkube.Client):
    # test dry_run parameter
    req_dry = respx.delete("https://localhost:9443/api/v1/nodes?dryRun=All")
    noeds = client.deletecollection(Node, dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'

    respx.delete("https://localhost:9443/api/v1/nodes")
    client.deletecollection(Node)


@respx.mock
def test_errors(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/xx").respond(content="Error", status_code=409)
    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/xx").respond(json={'message': 'got problems'},
              status_code=409)
    with pytest.raises(httpx.HTTPError):
        client.get(Pod, name="xx")

    with pytest.raises(lightkube.ApiError, match='got problems') as exc:
        client.get(Pod, name="xx")
    assert exc.value.status.message == 'got problems'


def make_watch_list(count=10):
    resp = "\n".join(
        json.dumps({'type': 'ADDED', 'object': {'metadata': {'name': f'p{i}', 'resourceVersion': '1'}}}) for i in
        range(count))
    return resp+"\n"


@respx.mock
def test_watch(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/nodes?watch=true").respond(content=make_watch_list())
    respx.get("https://localhost:9443/api/v1/nodes?watch=true&resourceVersion=1").respond(status_code=404)

    i = None
    with pytest.raises(httpx.HTTPError) as exi:
        for i, (op, node) in enumerate(client.watch(Node)):
            assert node.metadata.name == f'p{i}'
            assert op == 'ADDED'
    assert i == 9
    assert exi.value.response.status_code == 404


@respx.mock
def test_watch_version(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/nodes?resourceVersion=2&watch=true").respond(content=make_watch_list())
    respx.get("https://localhost:9443/api/v1/nodes?resourceVersion=1&watch=true").respond(status_code=404)

    # testing starting from specific resource version
    i = None
    with pytest.raises(httpx.HTTPError) as exi:
        for i, (op, node) in enumerate(client.watch(Node, resource_version="2")):
            assert node.metadata.name == f'p{i}'
            assert op == 'ADDED'
    assert i == 9
    assert exi.value.response.status_code == 404


@respx.mock
def test_watch_on_error(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/nodes?watch=true").respond(content=make_watch_list())
    respx.get("https://localhost:9443/api/v1/nodes?watch=true&resourceVersion=1").respond(status_code=404)

    i = None
    for i, (op, node) in enumerate(client.watch(Node, on_error=types.on_error_stop)):
        assert node.metadata.name == f'p{i}'
        assert op == 'ADDED'
    assert i == 9


@respx.mock
def test_watch_stop_iter(client: lightkube.Client):
    respx.get("https://localhost:9443/api/v1/nodes?watch=true").respond(content=make_watch_list())
    respx.get("https://localhost:9443/api/v1/nodes?watch=true&resourceVersion=1").respond(status_code=404)

    i = None
    for i, _ in enumerate(client.watch(Node, on_error=types.on_error_raise)):
        break
    assert i == 0


def make_wait_success():
    states = [
        {
            "type": "ADDED",
            "object": {
                "metadata": {"name": "test-node", "resourceVersion": "1"},
                "status": None,
            },
        },
        {
            "type": "ADDED",
            "object": {
                "metadata": {"name": "test-node", "resourceVersion": "1"},
                "status": {"conditions": [{"type": "TestCondition", "status": "True"}]},
            },
        },
    ]

    return "\n".join(map(json.dumps, states))


def make_wait_deleted():
    state = {
        "type": "DELETED",
        "object": {
            "metadata": {"name": "test-node", "resourceVersion": "1"},
            "status": {},
        },
    }

    return json.dumps(state)


def make_wait_failed():
    state = {
        "type": "ADDED",
        "object": {
            "metadata": {"name": "test-node", "resourceVersion": "1"},
            "status": {"conditions": [{"type": "TestCondition", "status": "True"}]},
        },
    }

    return json.dumps(state)


def make_wait_custom():
    state = {
        "type": "ADDED",
        "object": {
            "metadata": {"name": "custom-resource", "resourceVersion": "1"},
            "status": {"conditions": [{"type": "TestCondition", "status": "True"}]},
        },
    }

    return json.dumps(state)


@respx.mock
def test_wait_success(client: lightkube.Client):
    base_url = "https://localhost:9443/api/v1/nodes?fieldSelector=metadata.name%3Dtest-node&watch=true"

    respx.get(base_url).respond(content=make_wait_success())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_success())

    node = client.wait(Node, "test-node", for_conditions=["TestCondition"])

    assert node.to_dict()["metadata"]["name"] == "test-node"


@respx.mock
def test_wait_deleted(client: lightkube.Client):
    base_url = "https://localhost:9443/api/v1/nodes?fieldSelector=metadata.name%3Dtest-node&watch=true"

    respx.get(base_url).respond(content=make_wait_deleted())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_deleted())

    message = "nodes/test-node was unexpectedly deleted"
    with pytest.raises(lightkube.core.exceptions.ObjectDeleted, match=message):
        client.wait(Node, "test-node", for_conditions=["TestCondition"])


@respx.mock
def test_wait_failed(client: lightkube.Client):
    base_url = "https://localhost:9443/api/v1/nodes?fieldSelector=metadata.name%3Dtest-node&watch=true"

    respx.get(base_url).respond(content=make_wait_failed())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_failed())

    message = r"nodes/test-node has failure condition\(s\): TestCondition"
    with pytest.raises(lightkube.core.exceptions.ConditionError, match=message):
        client.wait(Node, "test-node", for_conditions=[], raise_for_conditions=["TestCondition"])


@respx.mock
def test_wait_custom(client: lightkube.Client):
    base_url = "https://localhost:9443/apis/custom.org/v1/customs?fieldSelector=metadata.name%3Dcustom-resource&watch=true"

    Custom = create_global_resource(
        group="custom.org", version="v1", kind="Custom", plural="customs"
    )
    respx.get(base_url).respond(content=make_wait_custom())
    respx.get(base_url + "&resourceVersion=1").respond(content=make_wait_custom())

    client.wait(Custom, "custom-resource", for_conditions=["TestCondition"])


@respx.mock
def test_patch_namespaced(client: lightkube.Client):
    # Default PatchType.STRATEGIC
    req = respx.patch("https://localhost:9443/api/v1/namespaces/default/pods/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = client.patch(Pod, "xx", Pod(metadata=ObjectMeta(labels={'l': 'ok'})))
    assert pod.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/strategic-merge-patch+json"

    # PatchType.MERGE
    req = respx.patch("https://localhost:9443/api/v1/namespaces/other/pods/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = client.patch(Pod, "xx", Pod(metadata=ObjectMeta(labels={'l': 'ok'})), namespace='other',
                       patch_type=types.PatchType.MERGE, force=True)
    assert pod.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/merge-patch+json"
    assert 'force' not in str(req.calls[0][0].url)  # force is ignored for non APPLY patch types

    # PatchType.APPLY
    req = respx.patch("https://localhost:9443/api/v1/namespaces/other/pods/xy?fieldManager=test").respond(
        json={'metadata': {'name': 'xy'}})
    pod = client.patch(Pod, "xy", Pod(metadata=ObjectMeta(labels={'l': 'ok'})), namespace='other',
                       patch_type=types.PatchType.APPLY, field_manager='test')
    assert pod.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # PatchType.APPLY + force
    req = respx.patch("https://localhost:9443/api/v1/namespaces/other/pods/xz?fieldManager=test&force=true").respond(
        json={'metadata': {'name': 'xz'}})
    pod = client.patch(Pod, "xz", Pod(metadata=ObjectMeta(labels={'l': 'ok'})), namespace='other',
                       patch_type=types.PatchType.APPLY, field_manager='test', force=True)
    assert pod.metadata.name == 'xz'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # PatchType.APPLY without field_manager
    with pytest.raises(ValueError, match="field_manager"):
        client.patch(Pod, "xz", Pod(metadata=ObjectMeta(labels={'l': 'ok'})), namespace='other',
                     patch_type=types.PatchType.APPLY)

    # test dry_run parameter
    req_dry = respx.patch("https://localhost:9443/api/v1/namespaces/other/pods/xz?fieldManager=test&dryRun=All").respond(
        json={'metadata': {'name': 'xz'}})
    node = client.patch(Pod, "xz", [{"op": "add", "path": "/metadata/labels/x", "value": "y"}],
        patch_type=types.PatchType.STRATEGIC, namespace="other",  field_manager='test', dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'


@respx.mock
def test_patch_global(client: lightkube.Client):
    req = respx.patch("https://localhost:9443/api/v1/nodes/xx").respond(json={'metadata': {'name': 'xx'}})
    node = client.patch(Node, "xx", [{"op": "add", "path": "/metadata/labels/x", "value": "y"}],
                        patch_type=types.PatchType.JSON)
    assert node.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/json-patch+json"

    # PatchType.APPLY + force
    req = respx.patch("https://localhost:9443/api/v1/nodes/xy?fieldManager=test&force=true").respond(
        json={'metadata': {'name': 'xy'}})
    node = client.patch(Node, "xy", Pod(metadata=ObjectMeta(labels={'l': 'ok'})),
                        patch_type=types.PatchType.APPLY, field_manager='test', force=True)
    assert node.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # test dry_run parameter
    req_dry = respx.patch("https://localhost:9443/api/v1/nodes/xz?fieldManager=test&dryRun=All").respond(
        json={'metadata': {'name': 'xz'}})
    node = client.patch(Node, "xz", [{"op": "add", "path": "/metadata/labels/x", "value": "y"}],
        patch_type=types.PatchType.APPLY, field_manager='test', dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'


@respx.mock
def test_field_manager(kubeconfig):
    config = KubeConfig.from_file(str(kubeconfig))
    client = lightkube.Client(config=config, field_manager='lightkube')
    respx.patch("https://localhost:9443/api/v1/nodes/xx?fieldManager=lightkube").respond(json={'metadata': {'name': 'xx'}})
    client.patch(Node, "xx", [{"op": "add", "path": "/metadata/labels/x", "value": "y"}],
                       patch_type=types.PatchType.JSON)

    respx.post("https://localhost:9443/api/v1/namespaces/default/pods?fieldManager=lightkube").respond(json={'metadata': {'name': 'xx'}})
    client.create(Pod(metadata=ObjectMeta(name="xx", labels={'l': 'ok'})))

    respx.put("https://localhost:9443/api/v1/namespaces/default/pods/xy?fieldManager=lightkube").respond(
        json={'metadata': {'name': 'xy'}})
    client.replace(Pod(metadata=ObjectMeta(name="xy")))

    respx.put("https://localhost:9443/api/v1/namespaces/default/pods/xy?fieldManager=override").respond(
        json={'metadata': {'name': 'xy'}})
    client.replace(Pod(metadata=ObjectMeta(name="xy")), field_manager='override')


@respx.mock
def test_create_namespaced(client: lightkube.Client):
    req = respx.post("https://localhost:9443/api/v1/namespaces/default/pods").respond(json={'metadata': {'name': 'xx'}})
    pod = client.create(Pod(metadata=ObjectMeta(name="xx", labels={'l': 'ok'})))
    json_contains(req.calls[0][0].read(), {"metadata": {"labels": {"l": "ok"}, "name": "xx"}})
    assert pod.metadata.name == 'xx'

    req2 = respx.post("https://localhost:9443/api/v1/namespaces/other/pods").respond(json={'metadata': {'name': 'yy'}})
    pod = client.create(Pod(metadata=ObjectMeta(name="xx", labels={'l': 'ok'})), namespace='other')
    assert pod.metadata.name == 'yy'
    json_contains(req2.calls[0][0].read(), {"metadata": {"labels": {"l": "ok"}, "name": "xx"}})

    respx.post("https://localhost:9443/api/v1/namespaces/ns2/pods").respond(
        json={'metadata': {'name': 'yy'}})
    pod = client.create(Pod(metadata=ObjectMeta(name="xx", labels={'l': 'ok'}, namespace='ns2')))
    assert pod.metadata.name == 'yy'

    # namespace inside object definition need to match with provided namespace parameter.
    with pytest.raises(ValueError):
        client.create(Pod(metadata=ObjectMeta(name="xx", namespace='ns1')), namespace='ns2')


@respx.mock
def test_create_global(client: lightkube.Client):
    req = respx.post("https://localhost:9443/api/v1/nodes").respond(json={'metadata': {'name': 'xx'}})
    pod = client.create(Node(metadata=ObjectMeta(name="xx")))
    json_contains(req.calls[0][0].read(), {"metadata": {"name": "xx"}})
    assert pod.metadata.name == 'xx'

    # dry-run
    req_dry = respx.post("https://localhost:9443/api/v1/nodes").respond(
        json={'metadata': {'name': 'xz'}})
    node = client.create(Node(metadata=ObjectMeta(name='xz')), dry_run=True)
    assert req_dry.calls[1][0].url.params['dryRun'] == 'All'

@respx.mock
def test_replace_namespaced(client: lightkube.Client):
    req = respx.put("https://localhost:9443/api/v1/namespaces/default/pods/xy").respond(json={'metadata': {'name': 'xy'}})
    pod = client.replace(Pod(metadata=ObjectMeta(name="xy")))
    json_contains(req.calls[0][0].read(), {"metadata": {"name": "xy"}})
    assert pod.metadata.name == 'xy'

    respx.put("https://localhost:9443/api/v1/namespaces/other/pods/xz").respond(json={'metadata': {'name': 'xz'}})
    pod = client.replace(Pod(metadata=ObjectMeta(name="xz")), namespace='other')
    assert pod.metadata.name == 'xz'

    # namespace inside object definition need to match with provided namespace parameter.
    with pytest.raises(ValueError):
        client.replace(Pod(metadata=ObjectMeta(name="xx", namespace='ns1')), namespace='ns2')

    # dry-run
    req_dry = respx.put("https://localhost:9443/api/v1/namespaces/other/pods/xx").respond(
        json={'metadata': {'name': 'xx'}})
    pod = client.replace(Pod(metadata=ObjectMeta(name='xx')), namespace="other", dry_run=True)
    assert pod.metadata.name == 'xx'
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'

@respx.mock
def test_replace_global(client: lightkube.Client):
    req = respx.put("https://localhost:9443/api/v1/nodes/xx").respond(json={'metadata': {'name': 'xx'}})
    pod = client.replace(Node(metadata=ObjectMeta(name="xx")))
    json_contains(req.calls[0][0].read(), {"metadata": {"name": "xx"}, "apiVersion": "v1", "kind": "Node"})
    assert pod.metadata.name == 'xx'

    # dry-run
    req_dry = respx.put("https://localhost:9443/api/v1/nodes/xy").respond(
        json={'metadata': {'name': 'xy'}})
    pod = client.replace(Node(metadata=ObjectMeta(name='xy')), dry_run=True)
    assert req_dry.calls[0][0].url.params['dryRun'] == 'All'


@respx.mock
def test_pod_log(client: lightkube.Client):
    result = ['line1\n', 'line2\n', 'line3\n']
    content = "".join(result)

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log").respond(content=content)
    lines = list(client.log('test'))
    assert lines == result

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log?follow=true").respond(
        content=content)
    lines = list(client.log('test', follow=True))
    assert lines == result

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log?tailLines=3").respond(
        content=content)
    lines = list(client.log('test', tail_lines=3))
    assert lines == result

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log?since=30&timestamps=true").respond(
        content=content)
    lines = list(client.log('test', since=30, timestamps=True))
    assert lines == result

    respx.get("https://localhost:9443/api/v1/namespaces/default/pods/test/log?container=bla").respond(
        content=content)

    lines = list(client.log('test', container="bla", newlines=False))
    assert lines == [_.strip() for _ in result]

    lines = list(client.log('test', container="bla"))
    assert lines == result


@respx.mock
def test_apply_namespaced(client: lightkube.Client):
    req = respx.patch("https://localhost:9443/api/v1/namespaces/default/pods/xy?fieldManager=test").respond(
        json={'metadata': {'name': 'xy'}})
    pod = client.apply(Pod(metadata=ObjectMeta(name='xy')), field_manager='test')
    assert pod.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # custom namespace, force
    req = respx.patch("https://localhost:9443/api/v1/namespaces/other/pods/xz?fieldManager=a&force=true").respond(
        json={'metadata': {'name': 'xz'}})
    pod = client.apply(Pod(metadata=ObjectMeta(name='xz', namespace='other')), field_manager='a', force=True)
    assert pod.metadata.name == 'xz'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # sub-resource
    req = respx.patch("https://localhost:9443/api/v1/namespaces/default/pods/xx/status?fieldManager=a").respond(
        json={'metadata': {'name': 'xx'}})
    pod = client.apply(Pod.Status(), name='xx', field_manager='a')
    assert pod.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"


@respx.mock
def test_apply_global(client: lightkube.Client):
    req = respx.patch("https://localhost:9443/api/v1/nodes/xy?fieldManager=test").respond(
        json={'metadata': {'name': 'xy'}})
    node = client.apply(Node(metadata=ObjectMeta(name='xy')), field_manager='test')
    assert node.metadata.name == 'xy'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # sub-resource + force
    req = respx.patch("https://localhost:9443/api/v1/nodes/xx/status?fieldManager=a&force=true").respond(
        json={'metadata': {'name': 'xx'}})
    node = client.apply(Node.Status(), name='xx', field_manager='a', force=True)
    assert node.metadata.name == 'xx'
    assert req.calls[0][0].headers['Content-Type'] == "application/apply-patch+yaml"

    # dry-run
    req = respx.patch("https://localhost:9443/api/v1/nodes/xz?fieldManager=test&dryRun=All").respond(
        json={'metadata': {'name': 'xz'}})
    node = client.apply(Node(metadata=ObjectMeta(name='xz')), field_manager='test', dry_run=True)
    assert node.metadata.name == 'xz'
    assert req.calls[0][0].url.params['dryRun'] == 'All'




"""This module provides a compatibility layer for functions in typing that appeared on 3.8"""

import collections.abc
import sys

if sys.version_info[:2] > (3, 7):
    from typing import get_args, get_origin

else:

    def get_origin(tp):
        if hasattr(tp, "__origin__"):
            return tp.__origin__
        return None

    def get_args(tp):
        if hasattr(tp, "__args__"):
            res = tp.__args__
            if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        return ()
